import sympy as sym
import numpy as np
from sympy import interpolating_spline


def bess_dcac(grid,data):
    '''

    PV+DCDC model MPPT control is implemented. 

    parameters
    ----------

    S_n: nominal power in VA
    U_n: nominal rms phase to phase voltage in V
    F_n: nominal frequency in Hz
    X_s: coupling reactance in pu (base machine S_n)
    R_s: coupling resistance in pu (base machine S_n)

    inputs
    ------

    p_s_ref: active power reference (pu, S_n base)
    q_s_ref: reactive power reference (pu, S_n base)
    v_dc: dc voltage in pu (when v_dc = 1 and m = 1, v_ac = 1)

    example
    -------

    "vscs": [{"bus":bus_name,"type":"pv_mpt_dcdc",
                 "S_n":1e6,"U_n":400.0,"F_n":50.0,
                 "X_s":0.1,"R_s":0.01,"monitor":True,
                 "I_sc":3.87,"V_oc":42.1,"I_mp":3.56,"V_mp":33.7,
                 "K_vt":-0.160,"K_it":0.065,
                 "N_pv_s":25,"N_pv_p":250}]
    
    '''
    
    bus_name = data['bus']
    name = bus_name
    bus_ac_name = bus_name

    # commons

    ## algebraic states
    p_dc = sym.Symbol(f"p_dc_{name}", real=True)

    ### HV-side
    R_h = sym.Symbol(f'R_h_{name}',real=True)
    p_r_ref = sym.Symbol(f'p_r_ref_{name}',real=True)
    m_h = sym.Symbol(f'm_h_{name}',real=True)
    i_g = sym.Symbol(f'i_g_{name}',real=True) 
    R_g = sym.Symbol(f'R_g_{name}',real=True)

    # dynamic states
    soc = sym.Symbol(f"soc_{name}", real=True)
    xi_soc = sym.Symbol(f"xi_soc_{name}", real=True)

    # algebraic states
    v_dc = sym.Symbol(f"v_dc_{name}", real=True)


    # Battery
    E_kWh = sym.Symbol(f"E_kWh_{name}", real=True)
    soc_min = sym.Symbol(f"soc_min_{name}", real=True)
    soc_max = sym.Symbol(f"soc_max_{name}", real=True)
    A_loss = sym.Symbol(f"A_loss_{name}", real=True)
    B_loss = sym.Symbol(f"B_loss_{name}", real=True)
    C_loss = sym.Symbol(f"C_loss_{name}", real=True)
    K_p = sym.Symbol(f"K_p_{name}", real=True)
    K_i = sym.Symbol(f"K_i_{name}", real=True)
    B_0 = sym.Symbol(f"B_0_{name}", real=True)
    B_1 = sym.Symbol(f"B_1_{name}", real=True)
    B_2 = sym.Symbol(f"B_2_{name}", real=True)
    B_3 = sym.Symbol(f"B_3_{name}", real=True)
    R_bat = sym.Symbol(f"R_bat_{name}", real=True)
    soc_ref = sym.Symbol(f"soc_ref_{name}", real=True) 

    epsilon_soc = soc_ref - soc
    p_soc = (K_p*epsilon_soc + K_i*xi_soc)
    E_n = 1000*3600*E_kWh

    if 'socs' in data:
        socs = np.array(data['socs'])
        es = np.array(data['es'])
        interpolation = interpolating_spline(1, soc, socs, es)
        interpolation._args = tuple(list(interpolation._args) + [sym.functions.elementary.piecewise.ExprCondPair(0,True)])
        e = interpolation
    else:
        e = B_0 + B_1*soc + B_2*soc**2 + B_3*soc**3
    i_dc = p_dc/v_dc

    ## dynamic equations    
    dsoc = 1/E_n*(-i_dc*e) #*0  + 1e-5*(soc_ref - soc)
    dxi_soc = epsilon_soc #-  1e-8*xi_soc  

    ## algebraic equations 
    #e = 867.0  
    g_v_dc = e - i_dc*R_bat - v_dc

    grid.dae['f'] += [dsoc, dxi_soc]
    grid.dae['x'] += [ soc,  xi_soc] 

    grid.dae['g'] +=     [g_v_dc]
    grid.dae['y_ini'] += [  v_dc]  
    grid.dae['y_run'] += [  v_dc]  

    soc_ref_N = 0.5
    if 'soc_0' in data:
        soc_ref_N = data['soc_0']

    grid.dae['u_ini_dict'].update({f'{str(soc_ref)}':soc_ref_N})
    grid.dae['u_run_dict'].update({f'{str(soc_ref)}':soc_ref_N})

    ## parameters  
    grid.dae['params_dict'].update({f"{K_p}":1e-4}) 
    grid.dae['params_dict'].update({f"{K_i}":1e-4})  
    grid.dae['params_dict'].update({f"{soc_min}":0.0}) 
    grid.dae['params_dict'].update({f"{soc_max}":1.0})           
    grid.dae['params_dict'].update({f"{E_kWh}":data['E_kWh']}) 

    if 'B_0' in data:
        B_0_N = data['B_0']
        B_1_N = data['B_1']
        B_2_N = data['B_2']
        B_3_N = data['B_3']
    else:
        B_0_N = 800
        B_1_N = 0.0
        B_2_N = 0.0
        B_3_N = 0.0

    grid.dae['params_dict'].update({f"{B_0}":B_0_N}) 
    grid.dae['params_dict'].update({f"{B_1}":B_1_N}) 
    grid.dae['params_dict'].update({f"{B_2}":B_2_N}) 
    grid.dae['params_dict'].update({f"{B_3}":B_3_N}) 

    if 'R_bat' in data:
        R_bat_N = data['R_bat']
    else:
        R_bat_N = 0.0

    grid.dae['params_dict'].update({f"{R_bat}":R_bat_N}) 

    grid.dae['xy_0_dict'].update({f"v_dc_{name}":800})


    # VSC

    ## inputs
    p_r_ref = sym.Symbol(f"p_r_ref_{name}", real=True)
    
    ## algebraic states
  

    p_ref = sym.Piecewise((p_r_ref,(p_r_ref <=0.0) & (soc<soc_max)),
                        (p_r_ref,(p_r_ref > 0.0) & (soc>soc_min)),
                        (0.0,True)) + p_soc
    
    ### AC-side
    p_a_ref,p_b_ref,p_c_ref = sym.symbols(f'p_vsc_a_ref_{bus_ac_name},p_vsc_b_ref_{bus_ac_name},p_vsc_c_ref_{bus_ac_name}',real=True)
    q_a_ref,q_b_ref,q_c_ref = sym.symbols(f'q_vsc_a_ref_{bus_ac_name},q_vsc_b_ref_{bus_ac_name},q_vsc_c_ref_{bus_ac_name}',real=True)
    p_ref,q_ref = sym.symbols(f'p_vsc_ref_{bus_ac_name},q_vsc_ref_{bus_ac_name}',real=True)

    p_ac,q_ac,p_loss = sym.symbols(f'p_vsc_{bus_ac_name},q_vsc_{bus_ac_name},p_vsc_loss_{bus_ac_name}',real=True)
    p_a_d,p_b_d,p_c_d,p_n_d = sym.symbols(f'p_a_d_{bus_ac_name},p_b_d_{bus_ac_name},p_c_d_{bus_ac_name},p_n_d_{bus_ac_name}',real=True)
   
    #### AC voltages:
    v_a_r,v_b_r,v_c_r,v_n_r = sym.symbols(f'V_{bus_ac_name}_0_r,V_{bus_ac_name}_1_r,V_{bus_ac_name}_2_r,V_{bus_ac_name}_3_r', real=True)
    v_a_i,v_b_i,v_c_i,v_n_i = sym.symbols(f'V_{bus_ac_name}_0_i,V_{bus_ac_name}_1_i,V_{bus_ac_name}_2_i,V_{bus_ac_name}_3_i', real=True)
   
    #### AC currents:
    i_a_r,i_a_i = sym.symbols(f'i_vsc_{bus_ac_name}_a_r,i_vsc_{bus_ac_name}_a_i',real=True)
    i_b_r,i_b_i = sym.symbols(f'i_vsc_{bus_ac_name}_b_r,i_vsc_{bus_ac_name}_b_i',real=True)
    i_c_r,i_c_i = sym.symbols(f'i_vsc_{bus_ac_name}_c_r,i_vsc_{bus_ac_name}_c_i',real=True)
    i_n_r,i_n_i = sym.symbols(f'i_vsc_{bus_ac_name}_n_r,i_vsc_{bus_ac_name}_n_i',real=True)

    # algebraic states dc side
    A_loss,B_loss,C_loss = sym.symbols(f'A_loss_{bus_ac_name},B_loss_{bus_ac_name},C_loss_{bus_ac_name}',real=True)     

    i_a_rms = sym.sqrt(i_a_r**2+i_a_i**2 + 1e-2) 
    i_b_rms = sym.sqrt(i_b_r**2+i_b_i**2 + 1e-2) 
    i_c_rms = sym.sqrt(i_c_r**2+i_c_i**2 + 1e-2) 
    i_n_rms = sym.sqrt(i_n_r**2+i_n_i**2 + 1e-2) 

    p_loss_a = C_loss + B_loss*i_a_rms + A_loss*i_a_rms*i_a_rms
    p_loss_b = C_loss + B_loss*i_b_rms + A_loss*i_b_rms*i_b_rms
    p_loss_c = C_loss + B_loss*i_c_rms + A_loss*i_c_rms*i_c_rms
    p_loss_n = C_loss + B_loss*i_n_rms + A_loss*i_n_rms*i_n_rms

    v_a = v_a_r + 1j*v_a_i
    v_b = v_b_r + 1j*v_b_i
    v_c = v_c_r + 1j*v_c_i
    v_n = v_n_r + 1j*v_n_i

    i_a = i_a_r + 1j*i_a_i
    i_b = i_b_r + 1j*i_b_i
    i_c = i_c_r + 1j*i_c_i
    i_n = i_n_r + 1j*i_n_i

    s_a = (v_a - v_n) * sym.conjugate(i_a)
    s_b = (v_b - v_n) * sym.conjugate(i_b)
    s_c = (v_c - v_n) * sym.conjugate(i_c)
    s_n = (v_n) * sym.conjugate(i_n)   

    p_a = p_a_ref + p_ref/3 + p_soc/3
    p_b = p_b_ref + p_ref/3 + p_soc/3
    p_c = p_c_ref + p_ref/3 + p_soc/3
    q_a = q_a_ref + q_ref/3
    q_b = q_b_ref + q_ref/3
    q_c = q_c_ref + q_ref/3

    # from powers to currents:
    eq_i_a_r =  sym.re(s_a) - p_a
    eq_i_b_r =  sym.re(s_b) - p_b
    eq_i_c_r =  sym.re(s_c) - p_c
    eq_i_a_i =  sym.im(s_a) - q_a
    eq_i_b_i =  sym.im(s_b) - q_b
    eq_i_c_i =  sym.im(s_c) - q_c
    
    eq_i_n_r = i_n_r + i_a_r + i_b_r + i_c_r
    eq_i_n_i = i_n_i + i_a_i + i_b_i + i_c_i
    
    p_ac = p_a + p_b + p_c
    p_loss_total = p_loss_a+p_loss_b+p_loss_c+p_loss_n 

    eq_p_dc = -p_dc +  p_ac + p_loss_total

    # eq_i_pos = -i_pos + p_dc/v_dc
    # eq_i_neg = -i_neg - p_dc/v_dc
    #eq_v_og = -v_og/R_gdc - i_pos - i_neg 


    grid.dae['g'] += [eq_i_a_r,eq_i_a_i,
                      eq_i_b_r,eq_i_b_i,
                      eq_i_c_r,eq_i_c_i,
                      eq_i_n_r,eq_i_n_i,
                     # eq_i_pos,eq_i_neg,#eq_v_og,
                      eq_p_dc
                      ]
    
    grid.dae['y_ini'] += [i_a_r,   i_a_i,
                          i_b_r,   i_b_i,
                          i_c_r,   i_c_i,
                          i_n_r,   i_n_i,
                          p_dc]

    grid.dae['y_run'] += [i_a_r,   i_a_i,
                          i_b_r,   i_b_i,
                          i_c_r,   i_c_i,
                          i_n_r,   i_n_i,
                          p_dc]

    # current injections ac side
    for ph in ['a','b','c','n']:
        i_s_r = sym.Symbol(f'i_vsc_{bus_ac_name}_{ph}_r', real=True)
        i_s_i = sym.Symbol(f'i_vsc_{bus_ac_name}_{ph}_i', real=True)  
        idx_r,idx_i = grid.node2idx(bus_ac_name,ph)
        grid.dae['g'] [idx_r] += -i_s_r
        grid.dae['g'] [idx_i] += -i_s_i
        i_s = i_s_r + 1j*i_s_i
        i_s_m = np.abs(i_s)
        grid.dae['h_dict'].update({f'i_vsc_{bus_ac_name}_{ph}_m':i_s_m})
    
      
    grid.dae['u_ini_dict'].update({f'{str(p_a_ref)}':0.0,f'{str(p_b_ref)}':0.0,f'{str(p_c_ref)}':0.0}) 
    grid.dae['u_ini_dict'].update({f'{str(q_a_ref)}':0.0,f'{str(q_b_ref)}':0.0,f'{str(q_c_ref)}':0.0}) 
    grid.dae['u_ini_dict'].update({f'{str(p_ref)}':0.0}) 
    grid.dae['u_ini_dict'].update({f'{str(q_ref)}':0.0}) 

    grid.dae['u_run_dict'].update({f'{str(p_a_ref)}':0.0,f'{str(p_b_ref)}':0.0,f'{str(p_c_ref)}':0.0}) 
    grid.dae['u_run_dict'].update({f'{str(q_a_ref)}':0.0,f'{str(q_b_ref)}':0.0,f'{str(q_c_ref)}':0.0}) 
    grid.dae['u_run_dict'].update({f'{str(p_ref)}':0.0}) 
    grid.dae['u_run_dict'].update({f'{str(q_ref)}':0.0}) 


    #grid.dae['u'].pop(str(v_dc_a_r))
    grid.dae['params_dict'].update({f'A_loss_{bus_ac_name}':0.0,f'B_loss_{bus_ac_name}':0.0,f'C_loss_{bus_ac_name}':0.0})
    
    grid.dae['h_dict'].update({f'p_vsc_{bus_ac_name}':sym.re(s_a)+sym.re(s_b)+sym.re(s_c)+sym.re(s_n)})
    grid.dae['h_dict'].update({f'p_vsc_loss_{bus_ac_name}':(p_loss_total)})
    grid.dae['h_dict'].update({f'v_dc_{bus_ac_name}':v_dc})



def test():
    import numpy as np
    import sympy as sym
    import json
    from pydae.urisi.urisi_builder import urisi
    import pydae.build_cffi as db

    # grid = urisi('bess_dcac.hjson')
    # grid.uz_jacs = True
    # grid.construct('temp')
    # grid.compile('temp')

    import temp

    S_n = 100e3
    V_n = 400
    I_n = S_n/V_n
    Conduction_losses = 0.02*S_n # = A*I_n**2
    losses = 1
    A = Conduction_losses/(I_n**2)*losses
    B = 1*losses
    C = 0.02*S_n*losses
    model = temp.model()
    model.ini({'A_loss_A1':A,'B_loss_A1':B,'C_loss_A1':C,'R_bat_A1':20e-3,'soc_ref_A1':1.0,
               'p_vsc_ref_A1':0e3},'xy_0.json')
    #model.report_params()
    model.report_u()
    model.report_y()
    model.report_z()
    model.Dt = 1.0
    model.run(3000, {'p_vsc_ref_A1': 100e3})
    model.run(6000, {'p_vsc_ref_A1':-100e3})

    model.post()
    model.report_x()
    model.report_z()

    import matplotlib.pyplot as plt

    fig,axes = plt.subplots(nrows=2,ncols=2)

    # axes[0,0].plot(model.Time,model.get_values('p_h_D1'), label='p_h')
    axes[0,1].plot(model.Time,model.get_values('v_dc_A1'), label='v_dc')
    axes[1,0].plot(model.Time,model.get_values('soc_A1'), label='soc')

    fig.savefig('bess_dcac.svg')


    for ax in axes.flatten():
        ax.grid()
        ax.legend()
        ax.set_xlim(0,model.Time[-1])
    axes[1,0].set_xlabel('Time (s)')
    axes[1,1].set_xlabel('Time (s)')

if __name__ == '__main__':

    #development()
    test()






import numpy as np
import matplotlib.pyplot as plt
import ipywidgets as widgets
import pydae.svg_tools as st 
import pydae.build_cffi as db
from pydae.bmapu import bmapu_builder


import ipywidgets
plt.style.use('https://raw.githubusercontent.com/pydae/pydae/master/src/pydae/edashboards/presentation.mplstyle')
colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
plt.rcParams['font.family'] =  ['sans-serif']

class dashboard():
    
    def __init__(self):

        
        self.colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
        self.data = {
        "system":{"name":"sm_pmsm_iso","S_base":100e6, "K_p_agc":0.01,"K_i_agc":0.01,"K_xif":0.0},       
        "buses":[{"name":"1", "P_W":0.0,"Q_var":0.0,"U_kV":0.4},
                 {"name":"2", "P_W":0.0,"Q_var":0.0,"U_kV":0.4},
                 {"name":"3", "P_W":0.0,"Q_var":0.0,"U_kV":0.4},
                ],
        "lines":[
            {"bus_j":"1", "bus_k":"2", "X_pu":0.05,"R_pu":0.01,"Bs_pu":0,"S_mva":2.0},
            {"bus_j":"2", "bus_k":"3", "X_pu":0.05,"R_pu":0.01,"Bs_pu":0,"S_mva":10.0}
                 ],
        "wecs":[
            {"type":"pmsm_1","bus":"1","S_n":1e6,
                "H_t":4.0,"H_r":1.0, "w_tr":5.0, "d_tr":0.01,
                "R_m":0.01,"L_m":0.05,"Phi_m":1.0,
                "R_s":0.02,"X_s":0.05,
                "K_pdc":0.1,"C_dc":0.5}],
        "syns":[
                {"bus":"2","S_n":2e6,
                    "X_d":1.81,"X1d":0.3, "T1d0":8.0,    
                    "X_q":1.81,"X1q":0.55,"T1q0":0.4,  
                    "R_a":0.02,"X_l": 0.2, 
                    "H":5.0,"D":1.0,
                    "Omega_b":314.1592653589793,"omega_s":1.0,"K_sec":1.0,
                    "avr":{"type":"sexs","K_a":100.0,"T_a":0.1,"T_b":0.1,"T_e":0.1,"E_min":-10.0,"E_max":10.0,"v_ref":1.0},
                    "gov":{"type":"tgov1","Droop":0.05,"T_1":1.0,"T_2":1.0,"T_3":1.0,"D_t":0.0,"p_c":0.0,"K_sec":1.0},
                    "K_delta":0.01}]
        }

    def ini(self):

        import sm_pmsm_iso

        self.model = sm_pmsm_iso.model()
        self.model.Dt = 0.01
        self.model.decimation = 1

        nu_w =11.0
        omega_0 = np.clip(nu_w/10,0.4,1.2)
        xy_0 = {
            "V_1": 1.0,
            "theta_1": 0.0,
            "V_2": 1.0,
            "theta_2": 0.0,
            "V_3": 1.0,
            "theta_3": 0.0,
            "omega_coi": 1.0,
            "omega_3": 1.0,
            "theta_tr_1": 0.0,
            "omega_t_1": omega_0,
            "omega_r_1": omega_0,
            "xi_beta_1":1,
            "beta_1":0.0,
            "i_sr_1": 1,
            "v_dc_1": 1.5,
            "p_w_mppt_lpf_1":1.2,
            "v_mq_1":1.0,
            "v_tq_ref_1":1.0
        }

        self.P_l_0 = 2.3e6
        params = {'K_pdc_1':50,"C_dc_1":0.01,"nu_w_1":nu_w,"P_3":-self.P_l_0,'K_a_2':200,
                "K_p_beta_1":1,"K_i_beta_1":1.0,"T_beta_1":1.2,"p_ref_ext_1":-0.0} 
        params.update({'K_p_pll_1':10, 'K_i_pll_1':100.0, 'T_pll_1':0.1,
                    'K_f_1':0, 'K_h_1':0})


        self.model.ini(params,xy_0)
        
        self.model.run(20.0,{})
        self.model.post();
           
    def build(self):



        grid = bmapu_builder.bmapu(self.data)
        #grid.checker()
        grid.uz_jacs = True
        grid.verbose = False
        grid.build('sm_pmsm_iso')

    def widgets(self):
        model = self.model
        colors = self.colors

        plt.ioff()
        plt.clf()

        fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(9, 4))

        self.line_p_g_2 = axes[0,0].plot(model.Time,model.get_values('p_g_2')*model.get_value('S_n_2')/1e6, label='SM Power (MW)')
        self.line_p_m_2 = axes[0,0].plot(model.Time,model.get_values('p_m_2')*model.get_value('S_n_2')/1e6, label='SM Mech. Power (MW)')

        self.line_p_s_1 = axes[0,0].plot(model.Time,model.get_values('p_s_1')*model.get_value('S_n_1')/1e6, label='WECS Power (MW)')

        self.line_omega_2 = axes[1,0].plot(model.Time, model.get_values('omega_2')*50, label='System frequency (Hz)', color=colors[1])
 
        self.line_omega_t = axes[0,1].plot(model.Time, model.get_values('omega_t_1'), label='WECS Turbine Speed (pu)')
        self.line_p_w = axes[1,1].plot(model.Time, model.get_values('p_w_1'), label='WECS Turbine Power (pu)')

        axes[0,0].set_ylim((0,3))
        axes[1,0].set_ylim((0.95*50,1.02*50))
        axes[0,1].set_ylim((0.8,1.3))
        axes[1,1].set_ylim((0.0,1.0))

        axes[0,0].grid(True)
        axes[1,0].grid(True)
        axes[0,1].grid(True)
        axes[1,1].grid(True)
        axes[0,0].legend(loc='best')
        axes[1,0].legend(loc='best')
        axes[0,1].legend(loc='best')
        axes[1,1].legend(loc='best')

        axes[1,0].set_xlabel('Time (s)')  
        axes[1,1].set_xlabel('Time (s)') 

        #self.fig.canvas.draw_idle()
        fig.savefig('results.svg')

        self.s = st.svg(r"results.svg")

        self.html = ipywidgets.HTML(
            value= self.s.tostring(),
            placeholder='',
            description='',
        )
        
        #axes[0].set_title('Par en función de la velocidad')
        #axes[1].set_title('Corriente en función de la velocidad')


        self.sld_P_l = ipywidgets.FloatSlider(orientation='horizontal',description = "P<sub>l</sub>", 
                                        value=0, min=0.0,max= 1.0, 
                                        step=0.1)
        self.sld_Q_l = ipywidgets.FloatSlider(orientation='horizontal',description = "Q<sub>l</sub>", 
                                        value=0, min=-100.0,max= 100, 
                                        step=1)


        self.sld_K_f = ipywidgets.FloatSlider(orientation='horizontal',description = "K<sub>f</sub>", 
                                        value=model.get_value('K_f_1'), min=0.0,max= 30, 
                                        step=1)


        self.sld_K_h = ipywidgets.FloatSlider(orientation='horizontal',description = "K<sub>h</sub>", 
                                        value=model.get_value('K_h_1'), min=0,max= 10, 
                                        step=1)

        self.fig = fig
        self.axes = axes

    def update(self,change):
        
        model = self.model

        DP_l = self.sld_P_l.value
        DQ_l = self.sld_Q_l.value
        K_f_1 = self.sld_K_f.value
        K_h_1 = self.sld_K_h.value

        model.Dt = 0.01
        model.ini({'P_3':-self.P_l_0,'Q_3':-0.0,'K_f_1':K_f_1,'K_h_1':K_h_1},'xy_0.json')
        model.run( 1.0,{})
        model.run(10,{'P_3':-DP_l*1e6-self.P_l_0})
        model.Dt = 0.1
        model.run(20,{'P_3':-DP_l*1e6-self.P_l_0})

        model.post();


        self.line_p_g_2[0].set_data(model.Time,model.get_values('p_g_2')*model.get_value('S_n_2')/1e6)
        self.line_p_m_2[0].set_data(model.Time,model.get_values('p_m_2')*model.get_value('S_n_2')/1e6)

        self.line_p_s_1[0].set_data(model.Time,model.get_values('p_s_1')*model.get_value('S_n_1')/1e6)

        self.line_omega_2[0].set_data(model.Time, model.get_values('omega_2')*50)

        self.line_p_w[0].set_data(model.Time, model.get_values('p_w_1'))
        self.line_omega_t[0].set_data(model.Time, model.get_values('omega_t_1'))

        #self.fig.canvas.draw_idle()

        self.fig.savefig('results_1.svg')
        self.s = st.svg(r"results_1.svg")
        self.html.value = self.s.tostring()

    def show(self):

        self.sld_P_l.observe(self.update, names='value')
        #self.sld_Q_l.observe(self.update, names='value')

        self.sld_K_h.observe(self.update, names='value')
        self.sld_K_f.observe(self.update, names='value')

        self.update(0)
        layout_row1 = ipywidgets.HBox([self.html])
        PQ = ipywidgets.VBox([self.sld_P_l])
        Ks = ipywidgets.VBox([self.sld_K_f,self.sld_K_h])
        layout_row2 = ipywidgets.HBox([PQ,Ks])

        self.layout = ipywidgets.VBox([layout_row1,layout_row2])
        display(self.layout)
        

    
if __name__ == '__main__':
    from pydae.edashboards.wecs.sm_pmsm_iso.sm_pmsm_iso_module import dashboard
    db = dashboard()
    db.build()

    db.ini()
    db.widgets()
    db.show()
    db.update(0)
        


# -*- coding: utf-8 -*-
"""
Created on Thu August 10 23:52:55 2022

@author: jmmauricio
"""

import numpy as np
import sympy as sym
from sympy import interpolating_spline

def bess_pq(grid,name,bus_name,data_dict):
    """
    # auxiliar
    
    """

    # inputs
    V_s = sym.Symbol(f"V_{bus_name}", real=True)
    theta_s = sym.Symbol(f"theta_{bus_name}", real=True)
    soc_ref = sym.Symbol(f"soc_ref_{name}", real=True)
    p_s_ref = sym.Symbol(f"p_s_ref_{name}", real=True)
    q_s_ref = sym.Symbol(f"q_s_ref_{name}", real=True)

        
    # dynamic states
    soc = sym.Symbol(f"soc_{name}", real=True)
    xi_soc = sym.Symbol(f"xi_soc_{name}", real=True)

    # algebraic states
    p_dc = sym.Symbol(f"p_dc_{name}", real=True)
    i_dc = sym.Symbol(f"i_dc_{name}", real=True)
    v_dc = sym.Symbol(f"v_dc_{name}", real=True)

    # parameters
    S_n = sym.Symbol(f"S_n_{name}", real=True)
    E_kWh = sym.Symbol(f"E_kWh_{name}", real=True)
    soc_min = sym.Symbol(f"soc_min_{name}", real=True)
    soc_max = sym.Symbol(f"soc_max_{name}", real=True)
    A_loss = sym.Symbol(f"A_loss_{name}", real=True)
    B_loss = sym.Symbol(f"B_loss_{name}", real=True)
    C_loss = sym.Symbol(f"C_loss_{name}", real=True)
    K_p = sym.Symbol(f"K_p_{name}", real=True)
    K_i = sym.Symbol(f"K_i_{name}", real=True)

    if 'socs' in data_dict:
        socs = np.array(data_dict['socs'])
        es = np.array(data_dict['es'])
        e_min = np.min(es)  # minimun voltage
        e_max = np.max(es)  # maximum voltage
        e_soc_order = 1
        if 'e_soc_order' in data_dict:
            e_soc_order = data_dict['e_soc_order']
        interpolation = interpolating_spline(e_soc_order, soc, socs, es)
        interpolation._args = tuple(list(interpolation._args) + [sym.functions.elementary.piecewise.ExprCondPair(e_max,True)])
        e = interpolation
        soc_ref_N = data_dict['soc_ref']
        e_ini = np.interp(soc_ref_N,socs,es)
    else:
        e = B_0 + B_1*soc
        e_ini = data_dict['B_0']

    R_bat = sym.Symbol(f"R_bat_{name}", real=True)

    # auxiliar
    H = E_kWh*1000.0*3600/S_n
    epsilon = soc_ref - soc
    p_soc = -(K_p*epsilon + K_i*xi_soc)
    p_s = sym.Piecewise((p_s_ref,(p_s_ref <= 0.0) & (soc<soc_max)),
                        (p_s_ref,(p_s_ref > 0.0) & (soc>soc_min)),
                        (0.0,True)) + p_soc
    #p_s = p_s_ref + p_soc
    q_s = q_s_ref
    s_s = (p_s**2 + q_s**2)**0.5
    i_s = s_s/V_s
    p_loss = A_loss*i_s**2 + B_loss*i_s + C_loss

    # dynamic equations    
    dsoc = 1/H*(-i_dc*e)   
    dxi_soc = epsilon     

    # algebraic equations   
    g_p_dc = p_s + p_loss - p_dc
    g_i_dc = v_dc*i_dc - p_dc
    g_v_dc = e - i_dc*R_bat - v_dc

    # dae 
    grid.dae['f'] += [dsoc, dxi_soc]
    grid.dae['x'] += [ soc,  xi_soc]
    grid.dae['g'] += [g_p_dc,g_i_dc,g_v_dc]
    grid.dae['y_ini'] += [p_dc, i_dc, v_dc]  
    grid.dae['y_run'] += [p_dc, i_dc, v_dc]  
    
    grid.dae['u_ini_dict'].update({f'{p_s_ref}':0.0})
    grid.dae['u_run_dict'].update({f'{p_s_ref}':0.0})          
    grid.dae['u_ini_dict'].update({f'{q_s_ref}':0.0})
    grid.dae['u_run_dict'].update({f'{q_s_ref}':0.0})     
    grid.dae['u_ini_dict'].update({f'{soc_ref}':0.5})
    grid.dae['u_run_dict'].update({f'{soc_ref}':0.5})      

    # outputs
    grid.dae['h_dict'].update({f"p_loss_{name}":p_loss})
    grid.dae['h_dict'].update({f"i_s_{name}":i_s})
    grid.dae['h_dict'].update({f"e_{name}":e})
    grid.dae['h_dict'].update({f"i_dc_{name}":i_dc})
    grid.dae['h_dict'].update({f"p_s_{name}":p_s})    
    grid.dae['h_dict'].update({f"q_s_{name}":q_s})

    # parameters  
    grid.dae['params_dict'].update({f"{K_p}":1e-6}) 
    grid.dae['params_dict'].update({f"{K_i}":1e-6})  
    grid.dae['params_dict'].update({f"{soc_min}":0.0}) 
    grid.dae['params_dict'].update({f"{soc_max}":1.0})           
    grid.dae['params_dict'].update({f"{S_n}":data_dict['S_n']}) 
    grid.dae['params_dict'].update({f"{E_kWh}":data_dict['E_kWh']}) 
    if 'A_loss' in data_dict:
        A_loss_N = data_dict['A_loss']
        B_loss_N = data_dict['B_loss']
        C_loss_N = data_dict['C_loss']
    else:
        A_loss_N = 0.0001
        B_loss_N = 0.0
        C_loss_N = 0.0001

    grid.dae['params_dict'].update({f"{A_loss}":A_loss_N}) 
    grid.dae['params_dict'].update({f"{B_loss}":B_loss_N}) 
    grid.dae['params_dict'].update({f"{C_loss}":C_loss_N}) 

    if 'R_bat' in data_dict:
        R_bat_N = data_dict['R_bat']
    else:
        R_bat_N = 0.0

    grid.dae['params_dict'].update({f"R_bat_{name}":R_bat_N}) 

    grid.dae['params_dict'].update({f"{S_n}":data_dict['S_n']}) 
    grid.dae['params_dict'].update({f"{E_kWh}":data_dict['E_kWh']}) 

    grid.dae['xy_0_dict'].update({str(v_dc):1.2})
    grid.dae['xy_0_dict'].update({str(soc):0.5})
    grid.dae['xy_0_dict'].update({str(xi_soc):0.5})


    p_W   = p_s*S_n
    q_var = q_s*S_n

    return p_W,q_var

def test():
    
    from pydae.bmapu import bmapu_builder
    import pytest

    grid = bmapu_builder.bmapu('bess_pq.hjson')
    #grid.checker()
    grid.verbose = False 
    grid.build('temp')

    import temp

    model = temp.model()
    soc_ref = 0.5
    model.ini({'soc_ref_1':soc_ref},'xy_0.json')

    assert model.get_value('soc_1') == pytest.approx(soc_ref, rel=0.001)
    assert model.get_value('p_dc_1') == pytest.approx(0.0)


    ### run:
    model.Dt = 1.0
    model.run(1.0,{})
    model.run(0.5*3600,{'p_s_ref_1': 1.0}) # half an hour discharging
  
    assert model.get_value('soc_1') == pytest.approx(0.25, rel=0.05)

    model.run(1.0*3600,{'p_s_ref_1':-1.0}) # half an hour discharging
    model.post()

    assert model.get_value('soc_1') == pytest.approx(0.5, rel=0.05)



if __name__=='__main__':

    test()








"""
===========================
Lines models
===========================

"""

from pydae.urisi.lines.lines import change_line
from pydae.urisi.lines.lines import codes_matrix
from pydae.urisi.lines.lines import lines2p_idx,change_line_fast



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Apr 14 11:53:28 2018

@author: jmmauricio
hola
"""

import numpy as np
import sympy as sym
#from sympy.diffgeom import TensorProduct
from sympy.physics.quantum  import TensorProduct
import json



def data_processing(data):
    for line in data['lines']:
        if 'X' in line: 
            L = line['X']/(2*np.pi*data['system']['f_hz'])
            line.update({'L':L})
        if 'B' in line: 
            C = line['B']/(2*np.pi*data['system']['f_hz'])
            line.update({'C':C})  
            
        if 'X_km' in line: 
            L = line['km']*line['X_km']/(2*np.pi*data['system']['f_hz'])
            line.update({'L':L})

        if 'R_km' in line: 
            R = line['km']*line['R_km']
            line.update({'R':R})
            
        if 'B_muS_km' in line: 
            B = line['km'] * line['B_muS_km']*1e-6
            C = B/(2*np.pi*data['system']['f_hz'])
            line.update({'C':C})   
            
        if 'B_km' in line: 
            B = line['km'] * line['B_km']
            C = B/(2*np.pi*data['system']['f_hz'])
            line.update({'C':C})   
            
    for load in data['loads']:
        if 'I_max' not in load: 
            if 'kVA' in load:
                I_max = load['kVA']*1000/690
            load['I_max'] = I_max
        if 'T_i' not in line: 
            load['T_i'] = 0.01        
    return data

def grid2dae_dq(data_input, park_type='original',dq_name='DQ'):
    
    if dq_name == 'DQ':
        D_ = 'D'
        Q_ = 'Q'        

    if type(data_input) == str:
        json_file = data_input
        json_data = open(json_file).read().replace("'",'"')
        data = json.loads(json_data)
    elif type(data_input) == dict:
        data = data_input


    data_processing(data)
    model_type = data['system']['model_type']
    buses = data['buses']
    lines = data['lines']
    loads = data['loads']
    grid_formers = data['grid_formers']

    buses_list = [item['bus'] for item in buses]
    load_buses_list = [item['bus'] for item in loads]
    gformers_buses_list = [item['bus'] for item in grid_formers]

    params = {}
    u_grid = {}
    f_grid = []
    g_grid = []
    y_grid_list = []
    x_grid_list = []
    x_list = []
    y_list = []
    
    omega = sym.Symbol('omega', real=True)
    M = len(lines) # total number of branches
    N = len(buses) # total number of buses
    A = sym.Matrix.zeros(M,cols=N)

    i_l_list = []
    i_list = []
    v_list = []
    R_list = []
    L_list = []
    C_list = [0]*N

    itl = 0
    for line in lines:
        sub_name = f"{line['bus_j']}{line['bus_k']}"

        idx_bus_j = buses_list.index(line['bus_j'])
        idx_bus_k = buses_list.index(line['bus_k'])

        A[itl,idx_bus_j] =  1
        A[itl,idx_bus_k] = -1

        bus_j = line['bus_j']
        bus_k = line['bus_k']    

        R_ij = sym.Symbol(f'R_{sub_name}', real=True)
        L_ij = sym.Symbol(f'L_{sub_name}', real=True)

        R_list += [R_ij]*2
        L_list += [L_ij]*2

        i_l_d = sym.Symbol(f'i_l_{sub_name}_{D_}', real=True)
        i_l_q = sym.Symbol(f'i_l_{sub_name}_{Q_}', real=True)
        i_l_list += [i_l_d,i_l_q]

        C_ij = sym.Symbol(f'C_{sub_name}', real=True)
        C_list[idx_bus_j] += C_ij/2
        C_list[idx_bus_k] += C_ij/2

        # parameters
        R_name = f'R_{sub_name}'
        R_value = line['R']
        L_name = f'L_{sub_name}'
        L_value = line['L']
        C_name = f'C_{sub_name}'
        C_value = line['C']

        params.update({R_name:R_value,L_name:L_value,C_name:C_value})

        itl += 1


    C_e_list = []
    for item in C_list:
        C_e_list += [item]
        C_e_list += [item]

    for bus in buses:
        bus_name = bus['bus']
        v_d = sym.Symbol(f'v_{bus_name}_{D_}', real=True)
        v_q = sym.Symbol(f'v_{bus_name}_{Q_}', real=True)
        i_d = sym.Symbol(f'i_{bus_name}_{D_}', real=True)
        i_q = sym.Symbol(f'i_{bus_name}_{Q_}', real=True)

        v_list += [v_d,v_q]
        i_list += [i_d,i_q]


    i_l_dq = sym.Matrix(i_l_list)
    R_e = sym.Matrix.diag(R_list)
    L_e = sym.Matrix.diag(L_list)
    Omega_list = sym.Matrix([[0,omega],[-omega,0]])
    Omega_e_M = sym.Matrix.diag([sym.Matrix([[0,omega],[-omega,0]])]*M)
    Omega_e_N = sym.Matrix.diag([sym.Matrix([[0,omega],[-omega,0]])]*N)

    C_e = sym.Matrix.diag(C_e_list)

    v_dq = sym.Matrix(v_list)
    i_dq = sym.Matrix(i_list)
    def T(P):
        u = TensorProduct(sym.Matrix.eye(P),sym.Matrix([1,0]).T)
        l = TensorProduct(sym.Matrix.eye(P),sym.Matrix([0,1]).T)
        return sym.Matrix([u,l])

    A_e = T(M).inv() @ sym.Matrix.diag([A,A]) @ T(N)

    if park_type == 'fisix':
        di_l_dq =  (-(R_e + L_e @ Omega_e_M) @ i_l_dq + A_e @ v_dq)
        dv_dq =  (-C_e @ Omega_e_N @ v_dq - A_e.T @ i_l_dq + i_dq)
    if park_type == 'original':
        di_l_dq =  (-(R_e - L_e @ Omega_e_M) @ i_l_dq + A_e @ v_dq)
        dv_dq =  (C_e @ Omega_e_N @ v_dq - A_e.T @ i_l_dq + i_dq)

    if model_type == 'ode':
        f_grid += list(L_e.inv()*di_l_dq)
        f_grid += list(C_e.inv()*dv_dq)
        x_grid_list += list(i_l_dq)                    # items as sym.Symbol
        x_grid_list += list(v_dq)                      # items as sym.Symbol
        x_list = [str(item) for item in x_grid_list]   # items as str
        
        for gformer in grid_formers:
            bus = gformer['bus']
            idx_D = x_list.index(f'v_{bus}_D')
            f_grid.pop(idx_D)
            x_grid_list.pop(idx_D)
            x_list.pop(idx_D)
            u_grid.update({f'v_{bus}_D':gformer["V_phph"]*np.sqrt(2/3)*np.sin(np.deg2rad(gformer["deg"]))})
            idx_Q = x_list.index(f'v_{bus}_Q')
            f_grid.pop(idx_Q)
            x_grid_list.pop(idx_Q)
            x_list.pop(idx_Q)
            u_grid.update({f'v_{bus}_Q':gformer["V_phph"]*np.sqrt(2/3)*np.cos(np.deg2rad(gformer["deg"]))})
        
    if model_type == 'dae':
        f_grid += list(L_e.inv()*di_l_dq)
        g_grid += list(dv_dq)
        x_grid_list += list(i_l_dq)                   # items as sym.Symbol
        y_grid_list += list(v_dq)                     # items as sym.Symbol
        x_list = [str(item) for item in x_grid_list]  # items as str
        y_list = [str(item) for item in y_grid_list]  # items as str
        
        for gformer in grid_formers:
            bus = gformer['bus']
            idx_D = y_list.index(f'v_{bus}_D')
            g_grid.pop(idx_D)
            y_grid_list.pop(idx_D)
            y_list.pop(idx_D)
            u_grid.update({f'v_{bus}_D':gformer["V_phph"]*np.sqrt(2/3)*np.sin(np.deg2rad(gformer["deg"]))})
            idx_Q = y_list.index(f'v_{bus}_Q')
            g_grid.pop(idx_Q)
            y_grid_list.pop(idx_Q)
            y_list.pop(idx_Q)
            u_grid.update({f'v_{bus}_Q':gformer["V_phph"]*np.sqrt(2/3)*np.cos(np.deg2rad(gformer["deg"]))})
            
            
    if model_type == 'ae':
        g_grid += list(di_l_dq)
        g_grid += list(dv_dq)
        y_grid_list += list(i_l_dq)
        y_grid_list += list(v_dq)

        for gformer in grid_formers:
            N_i_branch = len(list(i_l_dq))
            idx_gformer = buses_list.index(gformer['bus'])
            y_grid_list[N_i_branch+2*idx_gformer] = i_list[2*idx_gformer]
            y_grid_list[N_i_branch+2*idx_gformer+1] = i_list[2*idx_gformer+1]

            bus_name = gformer['bus']
            phi = np.deg2rad(gformer['deg'])
            v_d = np.sin(phi)*gformer['V_phph']*np.sqrt(2/3)
            v_q = np.cos(phi)*gformer['V_phph']*np.sqrt(2/3)
            u_grid.update({f'v_{bus_name}_{D_}':v_d,f'v_{bus_name}_{Q_}':v_q})

    for load in loads:

        bus_name = load['bus']
        i_d_ref = sym.Symbol(f'i_{bus_name}_d_ref', real=True)
        i_q_ref = sym.Symbol(f'i_{bus_name}_q_ref', real=True)
        i_d = sym.Symbol(f'i_{bus_name}_{D_}', real=True)
        i_q = sym.Symbol(f'i_{bus_name}_{Q_}', real=True)
        T_i = sym.Symbol(f'T_i_{bus_name}', real=True)
        I_max = sym.Symbol(f'I_max_{bus_name}', real=True)

        p_ref = sym.Symbol(f'p_{bus_name}_ref', real=True)
        q_ref = sym.Symbol(f'q_{bus_name}_ref', real=True)
        v_d = sym.Symbol(f'v_{bus_name}_{D_}', real=True)
        v_q = sym.Symbol(f'v_{bus_name}_{Q_}', real=True)

        den = v_d**2 + v_q**2

        den_sat = sym.Piecewise((0.01,den<0.01),(1e12,den>1e12),(den,True))

        if park_type == 'original':
            g_d = -i_d_ref + 2/3*(-p_ref*v_d - q_ref*v_q)/den_sat 
            g_q = -i_q_ref - 2/3*( p_ref*v_q - q_ref*v_d)/den_sat
            y_d = i_d_ref
            y_q = i_q_ref    



        i_d_sat = sym.Piecewise((-I_max,i_d_ref<-I_max),(I_max,i_d_ref>I_max),(i_d_ref,True))
        i_q_sat = sym.Piecewise((-I_max,i_q_ref<-I_max),(I_max,i_q_ref>I_max),(i_q_ref,True))

        f_d = 1/0.01*(i_d_sat - i_d)
        f_q = 1/0.01*(i_q_sat - i_q)
        f_grid += [f_d,f_q]

        u_grid.update({f'T_i_{bus_name}':load['T_i']})
        u_grid.update({f'I_max_{bus_name}':load['I_max']})

        g_grid += [g_d,g_q]
        y_grid_list += [i_d_ref,i_q_ref]
        x_grid_list += [i_d,i_q]

        if "kVA" in load:
            phi = np.arccos(load["pf"])
            p = load['kVA']*1000*np.cos(phi)
            q = load['kVA']*1000*np.sin(phi)*np.sign(load["pf"])
        if "kW" in load:
            p = load['kW']*1000
            q = load['kvar']*1000
        u_grid.update({f'p_{bus_name}_ref':p,f'q_{bus_name}_ref':q})

    for bus in buses:
        if bus['bus'] not in load_buses_list+gformers_buses_list:
            bus_name = bus['bus']
            params.update({f'i_{bus_name}_{D_}':0.0,f'i_{bus_name}_{Q_}':0.0})

    f_hz = data['system']['f_hz']
    params.update({'omega':2*np.pi*f_hz})

    x_list = [str(item) for item in x_grid_list]
  
    return {'f':f_grid,'g':g_grid,
            'x':x_grid_list,'y':y_grid_list, 'x_list':x_list,
            'u':u_grid,'params':params,'v_list':v_list}

 

def dcgrid2dae(data_input):
    vscs = data_input['grid_formers']
    park_type='original'
    dq_name='DQ'
    
    xy_0_dict = {}

    if dq_name == 'DQ':
        D_ = 'D'
        Q_ = 'Q'        

    if type(data_input) == str:
        json_file = data_input
        json_data = open(json_file).read().replace("'",'"')
        data = json.loads(json_data)
    elif type(data_input) == dict:
        data = data_input

    data_processing(data)
    model_type = data['system']['model_type']
    buses = data['buses']
    lines = data['lines']
    loads = data['loads']
    grid_formers = data['grid_formers']

    buses_list = [item['bus'] for item in buses]
    load_buses_list = [item['bus'] for item in loads]
    gformers_buses_list = [item['bus'] for item in grid_formers]

    params = {}
    u_grid = {}
    f_grid = []
    g_grid = []
    y_grid_list = []
    x_grid_list = []
    x_list = []
    y_list = []
    h_dict = {}

    omega = sym.Symbol('omega', real=True)
    M = len(lines) # total number of branches
    N = len(buses) # total number of buses
    A_k = sym.Matrix.zeros(M,cols=N)

    i_line_list = []
    i_list = []
    v_list = []
    R_list = []
    L_list = []
    C_list = [0]*N

    itl = 0
    for line in lines:
        sub_name = f"{line['bus_j']}{line['bus_k']}"

        idx_bus_j = buses_list.index(line['bus_j'])
        idx_bus_k = buses_list.index(line['bus_k'])

        A_k[itl,idx_bus_j] =  1
        A_k[itl,idx_bus_k] = -1

        bus_j = line['bus_j']
        bus_k = line['bus_k']    

        R_ij = sym.Symbol(f'R_{sub_name}', real=True)
        L_ij = sym.Symbol(f'L_{sub_name}', real=True)

        R_list += [R_ij]
        L_list += [L_ij]

        i_line = sym.Symbol(f'i_l_{sub_name}', real=True)
        i_line_list += [i_line]

        C_ij = sym.Symbol(f'C_{sub_name}', real=True)
        C_list[idx_bus_j] += C_ij/2
        C_list[idx_bus_k] += C_ij/2

        # parameters
        R_name = f'R_{sub_name}'
        R_value = line['R']
        L_name = f'L_{sub_name}'
        L_value = line['L']
        C_name = f'C_{sub_name}'
        C_value = line['C']

        params.update({R_name:R_value,L_name:L_value,C_name:C_value})

        itl += 1


    C_e_list = []
    for item in C_list:
        C_e_list += [item]
        C_e_list += [item]

    for bus in buses:
        bus_name = bus['bus']
        v = sym.Symbol(f'v_{bus_name}', real=True)
        i = sym.Symbol(f'i_{bus_name}', real=True)

        v_list += [v]
        i_list += [i]

        xy_0_dict.update({f'v_{bus_name}':3000})

    i_line = sym.Matrix(i_line_list)
    R_e = sym.Matrix.diag(R_list)
    L_e = sym.Matrix.diag(L_list)

    v  = sym.Matrix(v_list)
    i  = sym.Matrix(i_list)
    def T(P):
        u = TensorProduct(sym.Matrix.eye(P),sym.Matrix([1,0]).T)
        l = TensorProduct(sym.Matrix.eye(P),sym.Matrix([0,1]).T)
        return sym.Matrix([u,l])

    A_e = sym.Matrix.diag([A_k])

    if park_type == 'fisix':
        di_l_dq =  (-R_e  @ i_line + A_e @ v)
        dv =  ( - A_e.T @ i_line + i)
    if park_type == 'original':
        di_line =  (-(R_e) @ i_line + A_e @ v)
        dv =  (- A_e.T @ i_line + i)

    if model_type == 'ae':
        g_grid += list(di_line)
        g_grid += list(dv)
        y_grid_list += list(i_line_list)
        y_grid_list += list(v)

        for gformer in grid_formers:
            N_i_branch = len(list(i_line_list))
            idx_gformer = buses_list.index(gformer['bus'])
            y_grid_list[N_i_branch+idx_gformer] = i_list[idx_gformer]

            bus_name = gformer['bus']
            phi = np.deg2rad(gformer['deg'])
            v_d = np.sin(phi)*gformer['V_phph']*np.sqrt(2/3)
            v_q = np.cos(phi)*gformer['V_phph']*np.sqrt(2/3)
            u_grid.update({f'v_{bus_name}_{D_}':v_d,f'v_{bus_name}_{Q_}':v_q})
        
    return {'f':f_grid,'g':g_grid,
            'x':x_grid_list,'y':y_grid_list, 'x_list':x_list,
            'u':u_grid,'params':params,'v_list':v_list,'xy_0_dict':xy_0_dict}

def vsg2dae(data,grid_dae):
    '''
    

    Parameters
    ----------
    data : TYPE
        DESCRIPTION.
    grid_dae : TYPE
        DESCRIPTION.

    Returns
    -------
    dict
        DESCRIPTION.

    '''
    
    sin = sym.sin
    cos = sym.cos
    sqrt = sym.sqrt 
    
    
    vsgs = data['vsgs']
    N_vsg = len(vsgs)
    
    # secondary control
    p_sec = {}
    q_sec = {}
    xi_f_sec,xi_v_sec = sym.symbols('xi_f_sec,xi_v_sec', real=True)
    K_f_sec,K_v_sec = sym.symbols('K_f_sec,K_v_sec', real=True)
    
    omega_coi_h = 0
    H_total = 0
    N_voltage = 0
    v_prom = 0
    for vsg in vsgs:
        name = vsg['name']
        
        omega_v_i  = sym.Symbol(f'omega_v_{name}', real=True)
        H = sym.Symbol(f'H_{name}', real=True)
        
        omega_coi_h += omega_v_i*H
        H_total += H 
        
        v_s_filt_i = sym.symbols(f'v_s_filt_{name}', real=True)
        N_voltage += 1
        v_prom += v_s_filt_i/N_vsg
    
    omega_coi = omega_coi_h/H_total
    dxi_f_sec = 1 - omega_coi
    dxi_v_sec = 1 - v_prom
    
    for vsg in vsgs:
        name = vsg['name']
        p_sec.update({f'{name}':K_f_sec*xi_f_sec/N_vsg})
        q_sec.update({f'{name}':K_v_sec*xi_v_sec/N_vsg})

    f_ctrl_5 = [dxi_f_sec,dxi_v_sec]
    x_ctrl_5 = [ xi_f_sec, xi_v_sec]
    g_ctrl_5 = []
    y_ctrl_5 = []
    x_0_ctrl_5 = []
    y_0_ctrl_5 = []
    params_ctrl_5 = {'K_f_sec':0.001,'K_v_sec':0.01}
    u_ctrl_5 = {}
    h_ctrl_5 = {}

    
    f_vsg = []
    x_vsg = []
    g_vsg = []
    y_vsg = []
    x_0_vsg = []
    y_0_vsg = []
    params_vsg = {}
    u_vsg = {}
    h_vsg = {}

    for vsg in vsgs:
        name = vsg['name']
        bus = vsg['bus']

        U_b = vsg['U_b']
        S_b = vsg['S_b_kVA']*1000
        I_b = S_b/(np.sqrt(3)*U_b)

        U_bdq = U_b*(np.sqrt(2))
        V_bdq = U_bdq/np.sqrt(3)
        I_bdq = I_b*np.sqrt(2)




        ## Transformations #########################################################################
        ## feedbacks:
        feedbacks = ['i_tD','i_tQ'] + ['v_mD','v_mQ'] + ['i_sD','i_sQ'] + ['v_sD','v_sQ'] + ['phi']
        for item in feedbacks:
            exec(f"{item} = sym.Symbol('{item}_{name}', real=True)",globals())    



        v_md,v_mq = sym.symbols(f'v_md_{name},v_mq_{name}', real=True)
        v_sd,v_sq = sym.symbols(f'v_sd_{name},v_sq_{name}', real=True)
        i_sd,i_sq = sym.symbols(f'i_sd_{name},i_sq_{name}', real=True)
        i_td,i_tq = sym.symbols(f'i_td_{name},i_tq_{name}', real=True)
        phi,dum= sym.symbols(f'phi_{name},dum_{name}', real=True)

        eq_v_md = -v_md + v_mD*cos(phi) + v_mQ*sin(phi) # original park
        eq_v_mq = -v_mq - v_mD*sin(phi) + v_mQ*cos(phi) # original park

        eq_v_sd = -v_sd + v_sD*cos(phi) + v_sQ*sin(phi) # original park
        eq_v_sq = -v_sq - v_sD*sin(phi) + v_sQ*cos(phi) # original park

        eq_i_sd = -i_sd + i_sD*cos(phi) + i_sQ*sin(phi) # original park
        eq_i_sq = -i_sq - i_sD*sin(phi) + i_sQ*cos(phi) # original park

        # jmm: medimos i_t?
        eq_i_td = -i_td + i_tD*cos(phi) + i_tQ*sin(phi) # original park
        eq_i_tq = -i_tq - i_tD*sin(phi) + i_tQ*cos(phi) # original park

        g_aux = [eq_v_md,eq_v_mq,eq_v_sd,eq_v_sq,eq_i_td,eq_i_tq,eq_i_sd,eq_i_sq]    
        y_aux = [   v_md,   v_mq,   v_sd,   v_sq,   i_td,   i_tq,   i_sd,   i_sq]
        y_0_aux = [  0.0,  V_bdq,    0.0,  V_bdq,       0,      0]    


        #v_sd = v_md
        #v_sq = v_mq


       # S_b_kVA,U_b = sym.symbols(f'S_b_kVA_{name},U_b_{name}', real=True) # params



        ## per unit

        i_sd_pu = i_sd/I_bdq; # input in SI that is coverted to pu
        i_sq_pu = i_sq/I_bdq; # input in SI that is coverted to pu
        v_sd_pu = v_sd/V_bdq; # input in SI that is coverted to pu
        v_sq_pu = v_sq/V_bdq; # input in SI that is coverted to pu
        i_td_pu = i_td/I_bdq; # input in SI that is coverted to pu
        i_tq_pu = i_tq/I_bdq; # input in SI that is coverted to pu



        ## PLL #########################################################################


        # CTRL4 #########################################################################
        ## parameters:
        params_ctrl_4 = {}
        for item in ['T_vpoi','K_vpoi','T_f','K_f']:
            params_ctrl_4.update({f'{item}_{name}':vsg[item]})
            exec(f"{item} = sym.Symbol('{item}_{name}', real=True)",globals())
        ## inputs:
        u_ctrl_4 = {}
        for item in ['v_s_ref','omega_ref','p_r','q_r']:
            u_ctrl_4.update({f'{item}_{name}':vsg[item]})
            exec(f"{item} = sym.Symbol('{item}_{name}', real=True)",globals())
        ## dynamic states:
        x_list_ctrl_4 = ['omega_v_filt','v_s_filt']
        for item in x_list_ctrl_4:
            exec(f"{item} = sym.Symbol('{item}_{name}', real=True)",globals())
        ## algebraic states
        y_list_ctrl_4 = ['p_m_ref','q_s_ref']
        for item in y_list_ctrl_4:
            exec(f"{item} = sym.Symbol('{item}_{name}', real=True)",globals())    
        ## feedbacks:
        feedbacks = ['omega_v']
        for item in feedbacks:
            exec(f"{item} = sym.Symbol('{item}_{name}', real=True)",globals())       

        domega_v_filt = 1/T_f*(omega_v - omega_v_filt)
        dv_s_filt = 1/T_vpoi*((v_sd_pu**2+v_sq_pu**2)**0.5 - v_s_filt)

        eq_p_m_ref = -p_m_ref + p_r + K_f*(omega_ref - omega_v_filt) + p_sec[name]   # PFR and secondary input 
        eq_q_s_ref = -q_s_ref + q_r + K_vpoi*(v_s_ref - v_s_filt)    + q_sec[name]


        # from derivatives to the integrator
        f_ctrl_4 = [domega_v_filt,dv_s_filt];
        x_ctrl_4 = [ omega_v_filt, v_s_filt];
        g_ctrl_4 = [ eq_p_m_ref, eq_q_s_ref] #eq_i_sd_ref, eq_i_sq_ref, ]
        y_ctrl_4 = [    p_m_ref,    q_s_ref] #   i_sd_ref,    i_sq_ref,    omega_v,    DV_sat];



        # CTRL3 #########################################################################


        if vsg['ctrl3'] == 'uvsg_i':

            ## parameters:
            params_ctrl_3 = {}
            for item in ['K_p','T_p','K_q','T_q','R_v','X_v','S_b_kVA','U_b','K_phi','H','D']:
                params_ctrl_3.update({f'{item}_{name}':vsg[item]})
                exec(f"{item} = sym.Symbol('{item}_{name}', real=True)",globals())
            ## inputs:
            u_ctrl_3 = {}
            for item in ['p_m_ref','q_s_ref']: # []
                u_ctrl_3.update({f'{item}_{name}':vsg[item]})
                exec(f"{item} = sym.Symbol('{item}_{name}', real=True)",globals())
            ## dynamic states:
            x_list_ctrl_3 = ['phi','omega_v','xi_q','omega_rads']
            for item in x_list_ctrl_3:
                exec(f"{item} = sym.Symbol('{item}_{name}', real=True)",globals())
            ## algebraic states
            y_list_ctrl_3 = ['DV_sat','p_s_pu','q_s_pu']
            for item in y_list_ctrl_3:
                exec(f"{item} = sym.Symbol('{item}_{name}', real=True)",globals())    
            ## feedbacks:   
            v_dc,dum = sym.symbols(f'v_dc_{name},dum_{name}', real=True)

            # equations:
            eq_omega_v = -omega_v + K_p*(epsilon_p + xi_p/T_p) + 1.0; 

            dphi =  Omega_b*(omega_v-1.0) - K_phi*phi;
            dxi_p = epsilon_p;
            dxi_q = epsilon_q; 
            domega_rads = 1.0/1.0*(Omega_b*omega_v - omega_rads);


        if vsg['ctrl3'] == 'sm2':

            ## parameters:
            params_ctrl_3 = {}
            for item in ['K_p','T_p','K_q','T_q','R_v','X_v','S_b_kVA','U_b','K_phi','H','D']:
                params_ctrl_3.update({f'{item}_{name}':vsg[item]})
                exec(f"{item} = sym.Symbol('{item}_{name}', real=True)",globals())
            ## inputs:
            u_ctrl_3 = {}
            for item in ['p_m_ref','q_s_ref']: # []
                u_ctrl_3.update({f'{item}_{name}':vsg[item]})
                exec(f"{item} = sym.Symbol('{item}_{name}', real=True)",globals())
            ## dynamic states:
            x_list_ctrl_3 = ['phi','omega_v','xi_q','omega_rads']
            for item in x_list_ctrl_3:
                exec(f"{item} = sym.Symbol('{item}_{name}', real=True)",globals())
            ## algebraic states
            y_list_ctrl_3 = ['DV_sat','p_s_pu','q_s_pu']
            for item in y_list_ctrl_3:
                exec(f"{item} = sym.Symbol('{item}_{name}', real=True)",globals())    
            ## feedbacks:   
            v_dc,dum = sym.symbols(f'v_dc_{name},dum_{name}', real=True)


            # equations:

            fault_flag = 0;
            e_0 = 1.0;    
            eq_p_s_pu = -p_s_pu + i_sd_pu*v_sd_pu + i_sq_pu*v_sq_pu; # pu
            eq_q_s_pu = -q_s_pu + i_sd_pu*v_sq_pu - i_sq_pu*v_sd_pu; # pu 

            # from the integrator to the states
            epsilon_p = (p_m_ref - p_s_pu)*(1.0-fault_flag);
            epsilon_q =  q_s_ref - q_s_pu;


            dphi =  Omega_b*(omega_v-omega_coi) - K_phi*phi;
            domega_v = 1/(2*H)*(p_m_ref - p_s_pu - D*(omega_v - 1.0))
            dxi_q = epsilon_q; 
            domega_rads = 1.0/1.0*(Omega_b*omega_v - omega_rads);

            DV =   K_q*(epsilon_q + xi_q/T_q); 

            eq_DV_sat = DV_sat - DV;
            #if DV_sat > 0.1 
            #    DV_sat = 0.1;
            #    dxi_q = 0.0;
            #end
            #if DV_sat < -0.1 
            #    DV_sat = -0.1;
            #    dxi_q = 0.0;
            #end   

            e = e_0 + DV_sat;   


            if (not 'ctrl1' in vsg) and (not 'ctrl2' in vsg):                      # CTRL3 over CTRL0
                v_t_d_pu = 0.0
                v_t_q_pu = e

                v_t_d = (v_t_d_pu - R_v*i_sd_pu + X_v*i_sq_pu)*V_bdq
                v_t_q = (v_t_q_pu - R_v*i_sq_pu - X_v*i_sd_pu)*V_bdq

                eta_d_ref = v_t_d/v_dc*2
                eta_q_ref = v_t_q/v_dc*2

                # from derivatives to the integrator
                f_ctrl_3 = [dphi,domega_v,dxi_q,domega_rads];
                x_ctrl_3 = [ phi, omega_v, xi_q, omega_rads];
                g_ctrl_3 = [ eq_DV_sat, eq_p_s_pu, eq_q_s_pu] #eq_i_sd_ref, eq_i_sq_ref, ]
                y_ctrl_3 = [    DV_sat,    p_s_pu,    q_s_pu] #   i_sd_ref,    i_sq_ref,    omega_v,    DV_sat];
                x_0_ctrl_3 = [ 0.0, 0.0, 0.0, 2*np.pi*50]
                y_0_ctrl_3 = [ ] #     0.0,       V_bdq,          1]  


            if ('ctrl1' in vsg) and (not 'ctrl2' in vsg):                          # CTRL3 over CTRL1
                i_sd_pu = -(X_v*e + R_v*v_sd_pu + X_v*v_sq_pu)/(R_v**2 + X_v**2)
                i_sq_pu = -(R_v*e + R_v*v_sq_pu - X_v*v_sd_pu)/(R_v**2 + X_v**2);
                eq_i_sd_ref = -i_sd_ref + i_sd_pu*I_bdq
                eq_i_sq_ref = -i_sq_ref + i_sq_pu*I_bdq

                # from derivatives to the integrator
                f_ctrl_3 = [dphi,domega_v,dxi_q,domega_rads];
                x_ctrl_3 = [ phi, omega_v, xi_q, omega_rads];
                g_ctrl_3 = [ eq_DV_sat, eq_p_s_pu, eq_q_s_pu, eq_i_sd_ref, eq_i_sq_ref]
                y_ctrl_3 = [    DV_sat,    p_s_pu,    q_s_pu,    i_sd_ref,    i_sq_ref];
                x_0_ctrl_3 = [ 0.0, 0.0, 0.0, 2*np.pi*50]
                y_0_ctrl_3 = [ ] #     0.0,       V_bdq,          1]     

            if ('ctrl1' in vsg) and ('ctrl2' in vsg):                          # CTRL3 over CTRL2
                v_sd_ref_pu =    - R_v*i_sd_pu + X_v*i_sq_pu; 
                v_sq_ref_pu = -e - R_v*i_sq_pu - X_v*i_sd_pu;
                eq_v_sd_ref = -v_sd_ref + v_sd_ref_pu*V_bdq
                eq_v_sq_ref = -v_sq_ref + v_sq_ref_pu*V_bdq

                # from derivatives to the integrator
                f_ctrl_3 = [dphi,domega_v,dxi_q,domega_rads];
                x_ctrl_3 = [ phi, omega_v, xi_q, omega_rads];
                g_ctrl_3 = [ eq_DV_sat, eq_p_s_pu, eq_q_s_pu, eq_v_sd_ref, eq_v_sq_ref]
                y_ctrl_3 = [    DV_sat,    p_s_pu,    q_s_pu,    v_sd_ref,    v_sq_ref];
                x_0_ctrl_3 = [ 0.0, 0.0, 0.0, 2*np.pi*50]
                y_0_ctrl_3 = [ ] #     0.0,       V_bdq,          1]              


        if vsg['ctrl3'] == 'droop_pq':

            ## parameters:
            params_ctrl_3 = {}
            for item in ['Omega_b','K_p','T_p','K_q','T_q','R_v','X_v','S_b_kVA','U_b','K_phi','H','D','K_omega','K_v','T_omega']:
                params_ctrl_3.update({f'{item}_{name}':vsg[item]})
                exec(f"{item} = sym.Symbol('{item}_{name}', real=True)",globals())
            ## inputs:
            u_ctrl_3 = {}
            for item in ['p_m_ref','q_s_ref','e_ref']: # []
                u_ctrl_3.update({f'{item}_{name}':vsg[item]})
                exec(f"{item} = sym.Symbol('{item}_{name}', real=True)",globals())
            ## dynamic states:
            x_list_ctrl_3 = ['phi','omega_v','xi_q','omega_rads']
            for item in x_list_ctrl_3:
                exec(f"{item} = sym.Symbol('{item}_{name}', real=True)",globals())
            ## algebraic states
            y_list_ctrl_3 = ['p_s_pu','q_s_pu']
            for item in y_list_ctrl_3:
                exec(f"{item} = sym.Symbol('{item}_{name}', real=True)",globals())    
            ## feedbacks:   
            v_dc,dum = sym.symbols(f'v_dc_{name},dum_{name}', real=True)


            # equations:

            fault_flag = 0;
            e_0 = 1.0;    
            eq_p_s_pu = -p_s_pu + i_sd_pu*v_sd_pu + i_sq_pu*v_sq_pu; # pu
            eq_q_s_pu = -q_s_pu + i_sd_pu*v_sq_pu - i_sq_pu*v_sd_pu; # pu 

            # from the integrator to the states
            epsilon_p = (p_m_ref - p_s_pu)*(1.0-fault_flag);
            epsilon_q =  q_s_ref - q_s_pu;

            omega_v_ref = omega_ref + K_omega*epsilon_p 
            e =   e_ref + K_v*epsilon_q

            dphi =  Omega_b*(omega_v-omega_coi) - K_phi*phi;
            domega_v = 1/T_omega*(omega_v_ref - omega_v)
            domega_rads = 1.0/1.0*(Omega_b*omega_v - omega_rads);


            if (not 'ctrl1' in vsg) and (not 'ctrl2' in vsg):                      # CTRL3 over CTRL0
                v_t_d_pu = 0.0
                v_t_q_pu = e

                v_t_d = (v_t_d_pu - R_v*i_sd_pu + X_v*i_sq_pu)*V_bdq
                v_t_q = (v_t_q_pu - R_v*i_sq_pu - X_v*i_sd_pu)*V_bdq

                eta_d_ref = v_t_d/v_dc*2
                eta_q_ref = v_t_q/v_dc*2

                # from derivatives to the integrator
                f_ctrl_3 = [dphi,domega_v,domega_rads];
                x_ctrl_3 = [ phi, omega_v, omega_rads];
                g_ctrl_3 = [ eq_p_s_pu, eq_q_s_pu] #eq_i_sd_ref, eq_i_sq_ref, ]
                y_ctrl_3 = [    p_s_pu,    q_s_pu] #   i_sd_ref,    i_sq_ref,    omega_v,    DV_sat];
                x_0_ctrl_3 = [ 0.0, 0.0, 0.0, 2*np.pi*50]
                y_0_ctrl_3 = [ ] #     0.0,       V_bdq,          1]  


            if ('ctrl1' in vsg) and (not 'ctrl2' in vsg):                          # CTRL3 over CTRL1
                i_sd_pu = -(X_v*e + R_v*v_sd_pu + X_v*v_sq_pu)/(R_v**2 + X_v**2)
                i_sq_pu = -(R_v*e + R_v*v_sq_pu - X_v*v_sd_pu)/(R_v**2 + X_v**2);
                eq_i_sd_ref = -i_sd_ref + i_sd_pu*I_bdq
                eq_i_sq_ref = -i_sq_ref + i_sq_pu*I_bdq

                # from derivatives to the integrator
                f_ctrl_3 = [dphi,domega_rads];
                x_ctrl_3 = [ phi, omega_rads];
                g_ctrl_3 = [ eq_DV_sat, eq_p_s_pu, eq_q_s_pu, eq_i_sd_ref, eq_i_sq_ref]
                y_ctrl_3 = [    DV_sat,    p_s_pu,    q_s_pu,    i_sd_ref,    i_sq_ref];
                x_0_ctrl_3 = [ 0.0, 0.0, 0.0, 2*np.pi*50]
                y_0_ctrl_3 = [ ] #     0.0,       V_bdq,          1]     

            if ('ctrl1' in vsg) and ('ctrl2' in vsg):                          # CTRL3 over CTRL2
                v_sd_ref_pu =    - R_v*i_sd_pu + X_v*i_sq_pu; 
                v_sq_ref_pu = -e - R_v*i_sq_pu - X_v*i_sd_pu;
                eq_v_sd_ref = -v_sd_ref + v_sd_ref_pu*V_bdq
                eq_v_sq_ref = -v_sq_ref + v_sq_ref_pu*V_bdq

                # from derivatives to the integrator
                f_ctrl_3 = [dphi,domega_v,dxi_q,domega_rads];
                x_ctrl_3 = [ phi, omega_v, xi_q, omega_rads];
                g_ctrl_3 = [ eq_DV_sat, eq_p_s_pu, eq_q_s_pu, eq_v_sd_ref, eq_v_sq_ref]
                y_ctrl_3 = [    DV_sat,    p_s_pu,    q_s_pu,    v_sd_ref,    v_sq_ref];
                x_0_ctrl_3 = [ 0.0, 0.0, 0.0, 2*np.pi*50]
                y_0_ctrl_3 = [ ] #     0.0,       V_bdq,          1]        

        if vsg['ctrl3'] == 'genape':
            dphi =  Omega_b*(omega_v-omega_coi) - K_phi*phi;
            domega_v = RoCoF/F_b;






        # CTRL0 #########################################################################

        ## inputs:
        params_ctrl_0 = {}
        u_ctrl_0 = {}
        for item in []: # ['eta_d_ref','eta_q_ref','phi']: #+['eta_D','eta_Q']:
            u_ctrl_0.update({f'{item}_{name}':vsg[item]})
            exec(f"{item} = sym.Symbol('{item}_{name}', real=True)",globals())
        ## algebraic states
        y_list_ctrl_0 = ['eta_d','eta_q'] + ['eta_D','eta_Q'] 
        for item in y_list_ctrl_0:
            exec(f"{item} = sym.Symbol('{item}_{name}', real=True)",globals())        

        eq_eta_d = eta_d - eta_d_ref
        eq_eta_q = eta_q - eta_q_ref

        eq_eta_D = -eta_D + eta_d*cos(phi) - eta_q*sin(phi) # comment for test 1 
        eq_eta_Q = -eta_Q + eta_d*sin(phi) + eta_q*cos(phi) # comment for test 1

        # from derivatives to the integrator
        f_ctrl_0 = [];
        x_ctrl_0 = [];
        g_ctrl_0 = [eq_eta_d,eq_eta_q,eq_eta_D,eq_eta_Q]
        y_ctrl_0 = [   eta_d,   eta_q,   eta_D,   eta_Q];
        x_0_ctrl_0 = [ ]
        y_0_ctrl_0 = [0.0,  0.8, 0.0,  0.8]    



        ## VSC and Filters #########################################################################

        if vsg['filter'] == 'L':
            ## parameters:
            params_vsc_filter = {}
            for item in ['L_t','R_t','omega']:
                params_vsc_filter.update({f'{item}_{name}':vsg[item]})
                exec(f"{item} = sym.Symbol('{item}_{name}', real=True)",globals())
            ## inputs:
            u_vsc_filter= {}
            for item in ['v_dc']: #+['eta_D','eta_Q']:
                u_vsc_filter.update({f'{item}_{name}':vsg[item]})
                exec(f"{item} = sym.Symbol('{item}_{name}', real=True)",globals())
            ## dynamic states:
            x_list_vsc_filter = []
            for item in x_list_vsc_filter:
                exec(f"{item} = sym.Symbol('{item}_{name}', real=True)",globals())
            ## algebraic states
            y_list_vsc_filter = ['i_tD','i_tQ'] + ['v_mD','v_mQ'] + ['i_sD','i_sQ'] + ['v_sD','v_sQ']
            for item in y_list_vsc_filter:
                exec(f"{item} = sym.Symbol('{item}_{name}', real=True)")    
            ## feedbacks:
            v_poiD,v_poiQ = sym.symbols(f'v_{bus}_D,v_{bus}_Q', real=True)
            i_poiD,i_poiQ = sym.symbols(f'i_{bus}_D,i_{bus}_Q', real=True)

            #eta_D = eta_D_ref # - Gv_in*(i_tD - i_sD)
            #eta_Q = eta_Q_ref #- Gv_in*(i_tQ - i_sQ)

            # LCL filter
            di_tD = 1/L_t*(eta_D/2*v_dc - R_t*i_tD + omega*L_t*i_tQ - v_mD)  
            di_tQ = 1/L_t*(eta_Q/2*v_dc - R_t*i_tQ - omega*L_t*i_tD - v_mQ) 
            dv_mD = 1/C_m*(i_tD + C_m*omega*v_mQ - G_d*v_mD - i_sD) 
            dv_mQ = 1/C_m*(i_tQ - C_m*omega*v_mD - G_d*v_mQ - i_sQ) 
            di_sD = 1/L_s*(v_mD - R_s*i_sD + omega*L_s*i_sQ - v_sD)  
            di_sQ = 1/L_s*(v_mQ - R_s*i_sQ - omega*L_s*i_sD - v_sQ) 

            # Grid interaction
            eq_i_poiD =  i_sD - i_poiD
            eq_i_poiQ =  i_sQ - i_poiQ   

            eq_v_sD =  v_sD - v_poiD
            eq_v_sQ =  v_sQ - v_poiQ    

            grid_dae['params'].pop(f'i_{bus}_D')
            grid_dae['params'].pop(f'i_{bus}_Q')

            # DAE
            f_vsc_filter = []
            x_vsc_filter = []
            g_vsc_filter = [di_tD, di_tQ, eq_i_poiD, eq_i_poiQ, eq_v_sD, eq_v_sQ]
            y_vsc_filter = [ i_tD,  i_tQ,    i_poiD,    i_poiQ,    v_sD,    v_sQ]    
            x_0_vsc_filter = [ ]
            y_0_vsc_filter = [ 0.0,  0.0,       0.0,         0,       0,   V_bdq]

        if vsg['filter'] == 'LCL':
            ## parameters:
            params_vsc_filter = {}
            for item in ['L_t','R_t','C_m','L_s','R_s','omega','G_d']:
                params_vsc_filter.update({f'{item}_{name}':vsg[item]})
                exec(f"{item} = sym.Symbol('{item}_{name}', real=True)",globals())
            ## inputs:
            u_vsc_filter = {}
            for item in ['v_dc']: #+['eta_D','eta_Q']:
                u_vsc_filter.update({f'{item}_{name}':vsg[item]})
                exec(f"{item} = sym.Symbol('{item}_{name}', real=True)",globals())
            ## dynamic states:
            x_list_vsc_filter = []
            for item in x_list_vsc_filter:
                exec(f"{item} = sym.Symbol('{item}_{name}', real=True)")
            ## algebraic states
            y_list_vsc_filter = ['i_tD','i_tQ'] + ['v_mD','v_mQ'] + ['i_sD','i_sQ'] + ['v_sD','v_sQ']
            for item in y_list_vsc_filter:
                exec(f"{item} = sym.Symbol('{item}_{name}', real=True)",globals())    
            ## feedbacks:
            v_poiD,v_poiQ = sym.symbols(f'v_{bus}_D,v_{bus}_Q', real=True)
            i_poiD,i_poiQ = sym.symbols(f'i_{bus}_D,i_{bus}_Q', real=True)

            #eta_D = eta_D_ref # - Gv_in*(i_tD - i_sD)
            #eta_Q = eta_Q_ref #- Gv_in*(i_tQ - i_sQ)

            # LCL filter
            di_tD = 1/L_t*(eta_D/2*v_dc - R_t*i_tD + omega*L_t*i_tQ - v_mD)  
            di_tQ = 1/L_t*(eta_Q/2*v_dc - R_t*i_tQ - omega*L_t*i_tD - v_mQ) 
            dv_mD = 1/C_m*(i_tD + C_m*omega*v_mQ - G_d*v_mD - i_sD) 
            dv_mQ = 1/C_m*(i_tQ - C_m*omega*v_mD - G_d*v_mQ - i_sQ) 
            di_sD = 1/L_s*(v_mD - R_s*i_sD + omega*L_s*i_sQ - v_sD)  
            di_sQ = 1/L_s*(v_mQ - R_s*i_sQ - omega*L_s*i_sD - v_sQ) 

            # Grid interaction
            eq_i_poiD =  i_sD - i_poiD
            eq_i_poiQ =  i_sQ - i_poiQ   

            eq_v_sD =  v_sD - v_poiD
            eq_v_sQ =  v_sQ - v_poiQ    

            grid_dae['params'].pop(f'i_{bus}_D')
            grid_dae['params'].pop(f'i_{bus}_Q')

            # DAE
            f_vsc_filter = []
            x_vsc_filter = []
            g_vsc_filter = [di_tD, di_tQ, dv_mD, dv_mQ, di_sD, di_sQ, eq_i_poiD, eq_i_poiQ, eq_v_sD, eq_v_sQ]
            y_vsc_filter = [ i_tD,  i_tQ,  v_mD,  v_mQ,  i_sD,  i_sQ,    i_poiD,    i_poiQ,    v_sD,    v_sQ]    
            x_0_vsc_filter = [ ]
            y_0_vsc_filter = [ 0.0,  0.0,   0.0, V_bdq,     0,     0,         0,         0,       0,   V_bdq]


        ## Model integration
        f_vsg += f_vsc_filter + f_ctrl_0 + f_ctrl_3 + f_ctrl_4
        x_vsg += x_vsc_filter + x_ctrl_0 + x_ctrl_3 + x_ctrl_4
        g_vsg += g_vsc_filter + g_ctrl_0 + g_aux + g_ctrl_3 + g_ctrl_4  
        y_vsg += y_vsc_filter + y_ctrl_0 + y_aux + y_ctrl_3 + y_ctrl_4
        params_vsg.update(params_vsc_filter)
        params_vsg.update(params_ctrl_0)
        params_vsg.update(params_ctrl_3)
        params_vsg.update(params_ctrl_4)
        u_vsg.update(u_vsc_filter)
        u_vsg.update(u_ctrl_0)
        u_vsg.update(u_ctrl_3)
        u_vsg.update(u_ctrl_4)    

        h_vsg.update({f'i_sD_{name}':i_sD,f'i_sQ_{name}':i_sQ})

       # x_0_vsg += x_0_vsc_lc 
       # y_0_vsg += y_0_vsc_lc 

    f_vsg += f_ctrl_5
    x_vsg += x_ctrl_5
    params_vsg.update(params_ctrl_5)
    u_vsg.update(u_ctrl_5)
    
    return {'f_list':f_vsg,'g_list':g_vsg,
            'x_list':x_vsg,'y_list':y_vsg,
            'u_run_dict':u_vsg,'params_dict':params_vsg,'h_dict':h_vsg,
            'omega_coi':omega_coi}

def dcrail2dae(data_input,dcgrid_dae):
    h_dict = {}
    sections = data_input['sections']
    g_grid = dcgrid_dae['g']
    y_grid_list = dcgrid_dae['y']
    f_grid = dcgrid_dae['f'] 
    x_grid_list  = dcgrid_dae['x'] 

    for section in sections:

        nodes_i = section['nodes_i']

        for node in nodes_i:
            v = sym.Symbol(f'v_{node}', real=True)
            i = sym.Symbol(f'i_{node}', real=True)
            p = sym.Symbol(f'p_{node}', real=True)
            g_grid += [-p + v*i]
            y_grid_list += [i]

    for section in sections:

        nodes_v = section['nodes_v']

        for node in nodes_v:
            v = sym.Symbol(f'v_{node}', real=True)
            i = sym.Symbol(f'i_{node}', real=True)
            p = sym.Symbol(f'p_{node}', real=True)
            h_dict.update({f'p_{node}':v*i})
            h_dict.update({f'v_{node}':v})

    for section in sections[1:]:

        nodes_v = section['nodes_v']

        node = nodes_v[0]
        v_nom = sym.Symbol(f'v_nom', real=True)
        v = sym.Symbol(f'v_{node}', real=True)
        i = sym.Symbol(f'i_{node}', real=True)
        v_ref = sym.Symbol(f'v_ref_{node}', real=True)
        T_v = sym.Symbol(f'T_v', real=True)
        K_r = sym.Symbol(f'K_r', real=True)
        Dv_r = sym.Symbol(f'Dv_r_{node}', real=True)
        p = v*i
        v_ref = v_nom - K_r*p - Dv_r    # v_nom = nominal voltage, K_r*p: power droop, Dv_r remote input
        f_grid += [1/T_v*(v_ref-v)]  # gracias por el cambio
        x_grid_list += [v]
        
    dcgrid_dae.update({'h_dict':h_dict})

def pf_network(file_path):
    '''
    

    Parameters
    ----------
    file_path : string
        File path to the system data information.

    Returns
    -------
    dict
        Dictionary with the equations for pydae. 
        
    {
     'sys':{'name':'pf_1','S_base':100e6},       
     'buses':[{'name':'GRI','P_W':0.0,'Q_var':0.0,'U_kV':66.0, 'type':'slack'},
              {'name':'POI','P_W':0.0,'Q_var':0.0,'U_kV':66.0},
              {'name':'PMV','P_W':0.0,'Q_var':0.0,'U_kV':20.0}],
     'lines':[{'bus_j':'GRI','bus_k':'POI','X_km':0.4,'R_km':0.12,'km':20},
              {'bus_j':'POI','bus_k':'PMV','X_pu':0.04,'R_pu':0.01, 'S_mva':50.0}]
    }
        

    '''
    
    with open(file_path,'r') as fobj:
        data = json.loads(fobj.read().replace("'",'"'))
    sys = data['sys']
    buses = data['buses']
    lines = data['lines']

    params_grid = {'S_base':sys['S_base']}
    S_base = sym.Symbol("S_base", real=True) 
    N_bus = len(buses)
    N_branch = len(lines)
    A = sym.zeros(N_branch,N_bus)
    G_primitive = sym.zeros(N_branch,N_branch)
    B_primitive = sym.zeros(N_branch,N_branch)
    buses_list = [bus['name'] for bus in buses]
    it = 0
    for line in lines:

        bus_j = line['bus_j']
        bus_k = line['bus_k']

        idx_j = buses_list.index(bus_j)
        idx_k = buses_list.index(bus_k)    

        A[it,idx_j] = 1
        A[it,idx_k] =-1   

        line_name = f"{bus_j}_{bus_k}"
        g_jk = sym.Symbol(f"g_{line_name}", real=True) 
        b_jk = sym.Symbol(f"b_{line_name}", real=True) 
        G_primitive[it,it] = g_jk
        B_primitive[it,it] = b_jk

        if 'X_pu' in line:
            if 'S_mva' in line: S_line = 1e6*line['S_mva']
            R = line['R_pu']*sys['S_base']/S_line  # in pu of the system base
            X = line['X_pu']*sys['S_base']/S_line  # in pu of the system base
            G =  R/(R**2+X**2)
            B = -X/(R**2+X**2)
            params_grid.update({f"g_{line_name}":G})
            params_grid.update({f'b_{line_name}':B})

        if 'X' in line:
            bus_idx = buses_list.index(line['bus_j'])
            U_base = buses[bus_idx]['U_kV']
            Z_base = U_base**2/sys['S_base']
            R = line['R']/Z_base  # in pu of the system base
            X = line['X']/Z_base  # in pu of the system base
            G =  R/(R**2+X**2)
            B = -X/(R**2+X**2)
            params_grid.update({f"g_{line_name}":G})
            params_grid.update({f'b_{line_name}':B})

        if 'X_km' in line:
            bus_idx = buses_list.index(line['bus_j'])
            U_base = buses[bus_idx]['U_kV']*1000
            Z_base = U_base**2/sys['S_base']
            R = line['R_km']*line['km']/Z_base  # in pu of the system base
            X = line['X_km']*line['km']/Z_base  # in pu of the system base
            G =  R/(R**2+X**2)
            B = -X/(R**2+X**2)
            params_grid.update({f"g_{line_name}":G})
            params_grid.update({f'b_{line_name}':B})        

        it += 1


    G = A.T * G_primitive * A
    B = A.T * B_primitive * A    
    
    sin = sym.sin
    cos = sym.cos
    y_grid = []
    g = sym.zeros(2*N_bus,1)
    u_grid = {}
    h_grid = {}
    for j in range(N_bus):
        bus_j_name = buses_list[j]
        P_j = sym.Symbol(f"P_{bus_j_name}", real=True)
        Q_j = sym.Symbol(f"Q_{bus_j_name}", real=True)
        g[2*j]   = -P_j/S_base
        g[2*j+1] = -Q_j/S_base
        for k in range(N_bus): 

            bus_k_name = buses_list[k]
            V_j = sym.Symbol(f"V_{bus_j_name}", real=True) 
            V_k = sym.Symbol(f"V_{bus_k_name}", real=True) 
            theta_j = sym.Symbol(f"theta_{bus_j_name}", real=True) 
            theta_k = sym.Symbol(f"theta_{bus_k_name}", real=True) 
            g[2*j]   += V_j*V_k*(G[j,k]*cos(theta_j - theta_k) + B[j,k]*sin(theta_j - theta_k)) 
            g[2*j+1] += V_j*V_k*(G[j,k]*sin(theta_j - theta_k) - B[j,k]*cos(theta_j - theta_k))        
            h_grid.update({f"V_{bus_j_name}":V_j})
        bus = buses[j]
        bus_name = bus['name']
        if 'type' in bus:
            if bus['type'] == 'slack':
                y_grid += [P_j]
                y_grid += [Q_j]
                u_grid.update({f"V_{bus_name}":1.0})
                u_grid.update({f"theta_{bus_name}":0.0})  
        else:
            y_grid += [V_j]
            y_grid += [theta_j]        
            u_grid.update({f"P_{bus_name}":bus['P_W']})
            u_grid.update({f"Q_{bus_name}":bus['Q_var']})    
            
        params_grid.update({f'U_{bus_name}_n':bus['U_kV']*1000})
    g_grid = list(g)     

    if False:
        v_sym_list = []
        for bus in buses_list:
            V_m = sym.Symbol(f'V_{bus}',real=True)
            V_a = sym.Symbol(f'theta_{bus}',real=True)
            v_sym_list += [V_m*sym.exp(sym.I*V_a)]

        sym.Matrix(v_sym_list)

        I_lines = (G_primitive+1j*B_primitive) * A * sym.Matrix(v_sym_list)

        it = 0
        for line in lines:
            I_jk_r = sym.Symbol(f"I_{line['bus_j']}_{line['bus_k']}_r", real=True)
            I_jk_i = sym.Symbol(f"I_{line['bus_j']}_{line['bus_k']}_i", real=True)
            g_grid += [-I_jk_r + sym.re(I_lines[it])]
            g_grid += [-I_jk_i + sym.im(I_lines[it])]
            y_grid += [I_jk_r]
            y_grid += [I_jk_i]
            it += 1
            
    return {'g':g_grid,'y':y_grid,'u':u_grid,'h':h_grid, 'params':params_grid, 'data':data}




def syns_add(grid):
    sin = sym.sin
    cos = sym.cos
    buses = grid['data']['buses']
    buses_list = [bus['name'] for bus in buses]
    for syn in grid['data']['syns']:

        bus_name = syn['bus']
        idx_bus = buses_list.index(bus_name) # get the number of the bus where the syn is connected
        if not 'idx_powers' in buses[idx_bus]: buses[idx_bus].update({'idx_powers':0})
        buses[idx_bus]['idx_powers'] += 1

        P = sym.Symbol(f"P_{bus_name}_{buses[idx_bus]['idx_powers']}", real=True)
        Q = sym.Symbol(f"Q_{bus_name}_{buses[idx_bus]['idx_powers']}", real=True)
        V = sym.Symbol(f"V_{bus_name}", real=True)
        theta = sym.Symbol(f"theta_{bus_name}", real=True)
        i_d = sym.Symbol(f"i_d_{bus_name}", real=True)
        i_q = sym.Symbol(f"i_q_{bus_name}", real=True)
        delta = sym.Symbol(f"delta_{bus_name}", real=True)
        omega = sym.Symbol(f"omega_{bus_name}", real=True)
        p_m = sym.Symbol(f"p_m_{bus_name}", real=True)
        e1q = sym.Symbol(f"e1q_{bus_name}", real=True)
        e1d = sym.Symbol(f"e1d_{bus_name}", real=True)
        v_f = sym.Symbol(f"v_f_{bus_name}", real=True)
        v_c = sym.Symbol(f"v_c_{bus_name}", real=True)
        p_m_ref  = sym.Symbol(f"p_m_ref_{bus_name}", real=True)
        v_ref  = sym.Symbol(f"v_ref_{bus_name}", real=True)
        xi_m = sym.Symbol(f"xi_m_{bus_name}", real=True)
        
        v_d = V*sin(delta - theta) 
        v_q = V*cos(delta - theta) 

        for item in syn:
            string = f"{item}=sym.Symbol('{item}_{bus_name}', real = True)" 
            exec(string,globals())

        p_e = i_d*(v_d + R_a*i_d) + i_q*(v_q + R_a*i_q) 


        ddelta = Omega_b*(omega - omega_s) - K_delta*delta
        domega = 1/(2*H)*(p_m - p_e - D*(omega - omega_s))
        de1q = 1/T1d0*(-e1q - (X_d - X1d)*i_d + v_f)
        de1d = 1/T1q0*(-e1d + (X_q - X1q)*i_q)
        dv_c =   (V - v_c)/T_r
        dp_m =   (p_m_ref - p_m)/T_m
        dxi_m =   omega - 1

        g_id  = v_q + R_a*i_q + X1d*i_d - e1q
        g_iq  = v_d + R_a*i_d - X1q*i_q - e1d
        g_p  = i_d*v_d + i_q*v_q - P/S_n  
        g_q  = i_d*v_q - i_q*v_d - Q/S_n  
        g_vf  = K_a*(v_ref - v_c + v_pss) - v_f 
        g_pm  = -p_m_ref - K_sec*xi_m - 1/Droop*(omega - 1)

        f_syn = [ddelta,domega,de1q,de1d,dv_c,dp_m,dxi_m]
        x_syn = [ delta, omega, e1q, e1d, v_c, p_m, xi_m]
        g_syn = [g_id,g_iq,g_p,g_q,g_vf,g_pm]
        y_syn = [i_d,i_q,P,Q,v_f,p_m_ref]
        if 'f' not in grid: grid.update({'f':[]})
        if 'x' not in grid: grid.update({'x':[]})
        grid['f'] += f_syn
        grid['x'] += x_syn
        grid['g'] += g_syn
        grid['y'] += y_syn  
        
        S_base = sym.Symbol('S_base', real = True)
        grid['g'][idx_bus*2]   += -P/S_base
        grid['g'][idx_bus*2+1] += -Q/S_base
        
        for item in syn:       
            grid['params'].update({f"{item}_{bus_name}":syn[item]})
        grid['params'].pop(f"bus_{bus_name}")
        grid['params'].update({f"v_ref_{bus_name}":1.0})




def psys_builder(file_path):
    
    grid = pf_network(file_path)
    syns_add(grid)
    
    return grid


if __name__ == "__main__":
    file_path = './data/sys2bus.json'    
    grid = pf_network_shunt(file_path)



import json
import os

def read_data(data_input=''):
    
    if type(data_input) == str:
        if 'http' in data_input:
            url = data_input
            resp = requests.get(url)
            data = json.loads(resp.text)
        else:
            if os.path.splitext(data_input)[1] == '.json':
                with open(data_input,'r') as fobj:
                    data = json.loads(fobj.read().replace("'",'"'))
            if os.path.splitext(data_input)[1] == '.hjson':
                import hjson
                with open(data_input,'r') as fobj:
                    data = hjson.loads(fobj.read().replace("'",'"'))
    elif type(data_input) == dict:
        data = data_input

    return data


def save_json(data,file='json_out.json'):

    # Writing dictionary to a JSON file
    with open(file, 'w') as fobj:
        json.dump(data, fobj)

def save_hjson(data,file='json_out.json'):
    import hjson

    # Writing dictionary to a JSON file
    with open(file, 'w') as fobj:
        hjson.dump(data, fobj)
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Apr 14 11:53:28 2018

@author: jmmauricio

bug: if there is not jacobian add pass to the if mode == 12:
"""
import numpy as np
import sympy as sym
from collections import deque 
import pkgutil
import re
from sympy.matrices.sparsetools import _doktocsr
from sympy import SparseMatrix
from sympy.codegen.ast import Assignment
from scipy.sparse import csr_matrix,save_npz
import os


import cffi

import time
def sym_gen_str():

    str = '''\
import sympy as sym

if 'params' in sys_vars: params = sys_vars['params']
if 'params_dict' in sys_vars: params = sys_vars['params_dict']
u_list = sys_vars['u_list']
x_list = sys_vars['x_list']
y_list = sys_vars['y_list']
for item in params:
    exec(f"{item} = sym.Symbol('{item}')", globals())
for item in u_list:
    exec(f"{item} = sym.Symbol('{item}')", globals())
for item in x_list:
    exec(f"{item} = sym.Symbol('{item}')", globals())
for item in y_list:
    exec(f"{item} = sym.Symbol('{item}')", globals())  
sym_func_list = ['sin','cos']
for item in sym_func_list:
    exec(f"{item} = sym.{item}", globals()) 
    '''
    return str

def sym_gen(sys_vars):
    """Generates global sympy symbolic variables from a dictionary.
    
    Parameters
    ----------
    sys_vars : A ``dict`` with the following keys:
    
    params: ``dict``: Dictionary of the parameters and their values.
    u_list: ``list(str)``: List of input variables as strings for the running problem. 
    x_list: ``list(str)``: List of dynamical states as strings for the running problem.
    y_list: ``list(str)``: List of algebraic states as strings for the running problem.   

    R = 0.1
    L = 0.01
    T_pi = L/R
    K_p = L/0.01

    params= {'R' : 0.1, 'L':0.01, 'V_max':10.0, 'V_min':0.0, 'K_p':1.0, 'T_pi':0.01, 'K_aw':1.0}
    u_ini_dict = { 'i_ref':1.0}  # for the initialization problem
    u_run_dict = { 'i_ref':1.0}  # for the running problem (here initialization and running problem are the same)

    x_list = ['i','xi']
    y_ini_list = ['y','v']
    y_run_list = ['y','v']

    sys_vars = {'params':params,
                'u_list':u_run_dict,
                'x_list':x_list,
                'y_run_list':y_run_list}   
    """    

    params = sys_vars['params_dict']
    u_list = sys_vars['u_list']
    x_list = sys_vars['x_list']
    y_list = sys_vars['y_list']

    for item in params:
        exec(f"{item} = sym.Symbol('{item}', real=True)", globals())
    for item in u_list:
        exec(f"{item} = sym.Symbol('{item}', real=True)", globals())
    for item in x_list:
        exec(f"{item} = sym.Symbol('{item}', real=True)", globals())
    for item in y_list:
        exec(f"{item} = sym.Symbol('{item}', real=True)", globals())  
    sym_func_list = ['sin','cos']
    for item in sym_func_list:
        exec(f"{item} = sym.{item}", globals()) 


    return sys_vars

class builder():   

    def __init__(self,sys,verbose=False):

        self.verbose = verbose
        self.sys = sys
        self.jac_num = True
        self.name = self.sys['name']
        self.save_sources = False
        self.string_u2z = ''
        self.u2z = '\#'

        if not os.path.exists('build'):
            os.makedirs('build')

        self.matrices_folder = 'build'


    def check_system(self):
        sys = self.sys
        
        if len(sys['f_list']) == 0:
            print('system without dynamic equations, adding dummy dynamic equation')
            x_dummy,u_dummy = sym.symbols('x_dummy,u_dummy')
            sys['x_list'] = ['x_dummy']
            sys['f_list'] = [u_dummy-x_dummy]
            sys['u_ini_dict'].update({'u_dummy':1.0})
            sys['u_run_dict'].update({'u_dummy':1.0})

        if len(sys['g_list']) == 0:
            print('system without algebraic equations, adding dummy algebraic equation')
            y_dummy,u_dummy = sym.symbols('y_dummy,u_dummy')
            
            sys['g_list'] = [u_dummy-y_dummy]
            sys['y_ini_list'] = ['y_dummy']
            sys['y_run_list'] = ['y_dummy']
            sys['u_ini_dict'].update({'u_dummy':1.0})
            sys['u_run_dict'].update({'u_dummy':1.0})
            
        y_ini_set = set(sys['y_ini_list'])
        contains_duplicates = len(sys['y_ini_list']) != len(y_ini_set)
        if contains_duplicates:
            print('error: y_ini contains duplicates')

        y_set = set(sys['y_run_list'])
        contains_duplicates = len(sys['y_run_list']) != len(y_set)
        if contains_duplicates:
            print('error: y_run contains duplicates')
        
            
    def sym_jac(self,f,x):
        
        N_f = len(f)
        N_x = len(x)
        J = sym.zeros(N_f,N_x)
        
        for irow in range(N_f):
            str_f = str(f[irow])
            for icol in range(N_x):
                if str(x[icol]) in str_f:
                    J[irow,icol] = f[irow].diff(x[icol])
        
        return J

    def system(self):
        '''
        

        Parameters
        ----------
        sys : dict
            {'name':sys_name,
                'params_dict':params_dict,
                'f_list':f_list,
                'g_list':g_list,
                'x_list':x_list,
                'y_ini_list':y_ini_list,
                'y_run_list':y_run_list,
                'u_run_dict':u_run_dict,
                'u_ini_dict':u_ini_dict,
                'h_dict':h_dict}

        Returns
        -------
        sys : TYPE
            DESCRIPTION.

        '''

        sys = self.sys
        
        t_0 = time.time()
        if self.verbose: print(f'check_system (time: {time.time()-t_0})')
        self.check_system()
        
        f = sym.Matrix(sys['f_list']).T
        g = sym.Matrix(sys['g_list']).T
        x = sym.Matrix(sys['x_list']).T
        y_ini = sym.Matrix(sys['y_ini_list']).T
        y_run = sym.Matrix(sys['y_run_list']).T
        u_ini = sym.Matrix(list(sys['u_ini_dict'].keys()), real=True)
        
        u_run_list = [sym.Symbol(item,real=True) for item in list(sys['u_run_dict'].keys())]
        u_run = sym.Matrix(u_run_list).T 
        h =  sym.Matrix(list(sys['h_dict'].values())).T   
        

        if self.verbose: print(f'computing jacobians Fx_run,Fy_run  (time: {time.time()-t_0:0.3f} s)')
        Fx_run = f.jacobian(x)
        Fy_run = f.jacobian(y_run)
        if self.verbose: print(f'computing jacobians Gx_run,Gy_run  (time: {time.time()-t_0:0.3f} s)')
        Gx_run = g.jacobian(x)
        
        #Gy_run = g.jacobian(y_run): 
        Gy_run = self.sym_jac(g,y_run)

                

        if self.verbose: print(f'computing jacobians Fu_run,Gu_run  (time: {time.time()-t_0:0.3f} s)')
        Fu_run = f.jacobian(u_run)
    

        
        #Gu_run = g.jacobian(u_run):
        Gu_run = self.sym_jac(g,u_run)

        if self.verbose: print(f'computing jacobians Fx_ini,Fy_ini  (time: {time.time()-t_0:0.3f} s)')   
        Fx_ini = f.jacobian(x)
        Fy_ini = f.jacobian(y_ini)
        if self.verbose: print(f'computing jacobians Gx_ini,Gy_ini  (time: {time.time()-t_0:0.3f} s)')
        Gx_ini = g.jacobian(x)
        
        #Gy_ini = g.jacobian(y_ini)
        Gy_ini =  self.sym_jac(g,y_ini) 
                    
        if self.verbose: print(f'computing jacobians Hx_run,Hy_run,Hu_run  (time: {time.time()-t_0} s)')   
        Hx_run = h.jacobian(x)
        
        #Hy_run = h.jacobian(y_run) 
        Hy_run = self.sym_jac(h,y_run)
        
        
        #Hu_run = h.jacobian(u_run)
        Hu_run = self.sym_jac(h,u_run)


        sys['f']= f
        sys['g']= g 
        sys['x']= x
        sys['y_ini']= y_ini  
        sys['y_run']= y_run      
        sys['u_ini']= u_ini 
        sys['u_run']= u_run
        sys['h']= h 
        
        sys['Fx_run'] = Fx_run
        sys['Fy_run'] = Fy_run
        sys['Gx_run'] = Gx_run
        sys['Gy_run'] = Gy_run

        sys['Fx_ini'] = Fx_ini
        sys['Fy_ini'] = Fy_ini
        sys['Gx_ini'] = Gx_ini
        sys['Gy_ini'] = Gy_ini
        
        sys['Fu_run'] = Fu_run
        sys['Gu_run'] = Gu_run

        sys['Hx_run'] = Hx_run
        sys['Hy_run'] = Hy_run
        sys['Hu_run'] = Hu_run
        
        
        Fx_ini_rows,Fx_ini_cols = np.nonzero(Fx_ini)
        Fy_ini_rows,Fy_ini_cols = np.nonzero(Fy_ini)
        Gx_ini_rows,Gx_ini_cols = np.nonzero(Gx_ini)
        Gy_ini_rows,Gy_ini_cols = np.nonzero(Gy_ini)
        
        sys['Fx_ini_rows'] = Fx_ini_rows 
        sys['Fx_ini_cols'] = Fx_ini_cols 
        sys['Fy_ini_rows'] = Fy_ini_rows 
        sys['Fy_ini_cols'] = Fy_ini_cols 
        sys['Gx_ini_rows'] = Gx_ini_rows 
        sys['Gx_ini_cols'] = Gx_ini_cols 
        sys['Gy_ini_rows'] = Gy_ini_rows 
        sys['Gy_ini_cols'] = Gy_ini_cols 

        Fx_run_rows,Fx_run_cols = np.nonzero(Fx_run)
        Fy_run_rows,Fy_run_cols = np.nonzero(Fy_run)
        Gx_run_rows,Gx_run_cols = np.nonzero(Gx_run)
        Gy_run_rows,Gy_run_cols = np.nonzero(Gy_run)
        
        sys['Fx_run_rows'] = Fx_run_rows 
        sys['Fx_run_cols'] = Fx_run_cols 
        sys['Fy_run_rows'] = Fy_run_rows 
        sys['Fy_run_cols'] = Fy_run_cols 
        sys['Gx_run_rows'] = Gx_run_rows 
        sys['Gx_run_cols'] = Gx_run_cols 
        sys['Gy_run_rows'] = Gy_run_rows 
        sys['Gy_run_cols'] = Gy_run_cols 
        
        sys['N_x'] = len(x)
        sys['N_y'] = len(y_run)
        sys['N_u'] = len(u_run)
        sys['N_z'] = len(h)

        if self.verbose: print(f'end system  (time: {time.time()-t_0:0.3f} s)')   
        
        return sys


    def jacobians(self):
        '''
        

        Parameters
        ----------
        sys : dict
            

        Returns
        -------
        sys : TYPE
            DESCRIPTION.

        '''
        
        sys = self.sys
        
        t_0 = time.time()
        
        N_x = sys['N_x']
        N_y = sys['N_y']
        N_z = sys['N_z']
        N_u = sys['N_u']
    
        Fx_run = sys['Fx_run']
        Fy_run = sys['Fy_run']
        Gx_run = sys['Gx_run']
        Gy_run = sys['Gy_run']
        
        Fx_ini = sys['Fx_run']
        Fy_ini = sys['Fy_ini']
        Gx_ini = sys['Gx_ini']
        Gy_ini = sys['Gy_ini']

        Fu_run = sys['Fu_run']
        Gu_run = sys['Gu_run'] 
        
        Hx_run = sys['Hx_run'] 
        Hy_run = sys['Hy_run'] 
        Hu_run = sys['Hu_run']
        
        if self.verbose: print(f'computing jac_ini (time: {time.time()-t_0})')
        jac_ini = sym.Matrix([[Fx_ini,Fy_ini],[Gx_ini,Gy_ini]]) 
        
        if self.verbose: print(f'computing jac_run (time: {time.time()-t_0})')
        jac_run = sym.Matrix([[Fx_run,Fy_run],[Gx_run,Gy_run]]) 
        
        if self.verbose: print(f'computing jac_trap (time: {time.time()-t_0})')
        eye = sym.eye(N_x, real=True)
        Dt = sym.Symbol('Dt',real=True)
        jac_trap = sym.Matrix([[eye - 0.5*Dt*Fx_run, -0.5*Dt*Fy_run],[Gx_run,Gy_run]])    

        sys['jac_ini']  = jac_ini
        sys['jac_run']  = jac_run   
        sys['jac_trap'] = jac_trap
        

        if self.verbose: print(f'end of jacobians computation (time: {time.time()-t_0:0.3f})')



    def getIndex(self,s, i): 
    
        # If input is invalid. 
        if s[i] != '(':
            return -1
    
        # Create a deque to use it as a stack. 
        d = deque() 
    
        # Traverse through all elements 
        # starting from i. 
        for k in range(i, len(s)): 
    
            # Pop a starting bracket 
            # for every closing bracket 
            if s[k] == ')': 
                d.popleft() 
    
            # Push all starting brackets 
            elif s[k] == '(': 
                d.append(s[i]) 
    
            # If deque becomes empty 
            if not d: 
                return k 
    
        return -1
    
    def arg2np(self,string,function_name):
        idx_end = 0
        for it in range(3):
            if function_name in string[idx_end:]:
                #print(string[idx_end:])
                idx_ini = string.find(f'{function_name}(',idx_end)+len(f'{function_name}(')
                idx_end = self.getIndex(string, idx_ini-1)
                string = string.replace(string[idx_ini:idx_end],f'np.array([{string[idx_ini:idx_end]}])')
        else: pass
        return string    

    def replace(self,eq, x_list, y_list, u_list, params_list):
        eq_str = str(eq)
        it = 0
        for item in x_list:
            x_sym = sym.Symbol(f'x_{it:04}',real=True)
            eq = eq.replace(item,x_sym)
            eq_str = str(eq)
            eq_str = eq_str.replace(str(x_sym), f'x[{it}]')
            it +=1
        it = 0
        for item in y_list:
            y_sym = sym.Symbol(f'y_{it:04}',real=True)
            eq = eq.replace(item,y_sym)
            eq_str = str(eq)
            eq_str = eq_str.replace(str(y_sym), f'y[{it}]')
            it +=1
        it = 0
        for item in u_list:
            u_sym = sym.Symbol(f'u_{it:04}')
            eq = eq.subs(str(item),u_sym)
            eq_str = str(eq)
            eq_str = eq_str.replace(str(u_sym), f'u[{it}]')
            it +=1
        it = 0
        for item in params_list:
            item_sym = sym.Symbol(item)
            p_sym = sym.Symbol(f'p_{it:04}',real=True)
            eq = eq.replace(item_sym,p_sym)
            eq_str = str(eq)
            eq_str = eq_str.replace(str(p_sym), f'p[{it}]')
            it +=1    

        eq_str = str(eq)
        
        it = 0
        for item in x_list:
            x_sym = sym.Symbol(f'x_{it:04}',real=True)        
            eq_str = eq_str.replace(str(x_sym), f'x[{it}]')
            it +=1
        it = 0
        for item in y_list:
            y_sym = sym.Symbol(f'y_{it:04}',real=True)
            eq_str = eq_str.replace(str(y_sym), f'y[{it}]')
            it +=1
        it = 0
        for item in u_list:
            u_sym = sym.Symbol(f'u_{it:04}')
            eq_str = eq_str.replace(str(u_sym), f'u[{it}]')
            it +=1
        it = 0
        for item in params_list:
            item_sym = sym.Symbol(item)
            p_sym = sym.Symbol(f'p_{it:04}',real=True)
            eq_str = eq_str.replace(str(p_sym), f'p[{it}]')
            it +=1   
            
        return eq_str

    def mreplace(string,replacements):
        for item in replacements:
            string = str(string).replace(item,replacements[item])
        return string

    # def sym2str(sym_exp,x,y,u,p,multi_eval=False):
        
    #     matrix_str = str(sym_exp)
        
    #     mev = ''
    #     if multi_eval:
    #         mev = 'i,'
        
    #     if sym_exp == 0:
    #         matrix_str = 0
    #         return matrix_str
            

    #     for it in range(len(x)):
    #         name = str(x[it])
    #         matrix_str = re.sub(r'\b' + name + r'\b',f'x[{mev}{it}]',matrix_str)
    #     for it in range(len(y)):
    #         name = str(y[it])
    #         matrix_str = re.sub(r'\b' + name + r'\b',f'y[{mev}{it}]',matrix_str)
    #     for it in range(len(u)):
    #         name = str(u[it])
    #         matrix_str = re.sub(r'\b' + name + r'\b',f'u[{mev}{it}]',matrix_str)
    #     it = 0
    #     for item in p:
    #         name = str(item)
    #         matrix_str = re.sub(r'\b' + name + r'\b',f'p[{mev}{it}]',matrix_str)
    #         it+=1
            
    #     return matrix_str   

    def vector2string(self,vector,function_header,vector_name,x,y,u,p,aux={},multi_eval=False):
        string = ''
        string_i = ''
        N  = len(vector)
        
        for item in aux:
            str_element = self.sym2str(aux[item],x,y,u,p,multi_eval=multi_eval)
            string_i += f'{" "*4}{item} = {str_element}\n'
        
        string_i += '\n'
        
        for it in range(N):
            str_element = self.sym2str(vector[it],x,y,u,p,multi_eval=multi_eval)
            str_element = self.arg2np(str_element,"Piecewise")

            if multi_eval:
                string_i += f'{" "*4}{vector_name}[i,{it}] = {str_element}\n'
            else:
                string_i += f'{" "*4}{vector_name}[{it}] = {str_element}\n'            

        string += '\n'
        string += function_header
        string += string_i
        string += '\n'
        
        return string

    def matrix2string(self,matrix,function_header,matrix_name,x,y,u,p):
        string = ''
        string_xy = ''
        string_up = ''
        string_num = ''
        N_row,N_col = matrix.shape
        
        for irow in range(N_row):
            for icol in range(N_col):
                element = matrix[irow,icol]
                
                if element.is_number:
                    str_element = str(element)
                else:
                    str_element = self.sym2str(element,x,y,u,p)
                
                if str_element != '0':
                    str_element = self.arg2np(str_element,"Piecewise")
                    if 'x' in str_element or 'y' in str_element:
                        string_xy += f'{" "*4}{matrix_name}[{irow},{icol}] = {str_element}\n'
                    elif 'u' in str_element or 'p' in str_element or 'Dt' in str_element:
                        string_up += f'{" "*4}{matrix_name}[{irow},{icol}] = {str_element}\n'
                    else:
                        string_num += f'{" "*4}{matrix_name}[{irow},{icol}] = {str_element}\n'

                        

        string += '\n'
        string += '@numba.njit(cache=True)\n'
        string += function_header.replace('(','_xy(')
        string += string_xy
        string += '\n'
        string += '@numba.njit(cache=True)\n'
        string += function_header.replace('(','_up(')
        string += string_up
        string += '\n'
        string += function_header.replace('(','_num(')
        string += string_num
        string += '\n'
        
        return string

    def spmatrix2def(self,spmatrix_list,matrix_name):
        
        string = ''
        string += f'def {matrix_name}_vectors():\n\n'
        string += f'{" "*4}{matrix_name}_ia = ' + str(spmatrix_list[1]) + '\n'
        string += f'{" "*4}{matrix_name}_ja = ' + str(spmatrix_list[2]) + '\n'    
        string += f'{" "*4}{matrix_name}_nia = ' + str(spmatrix_list[3][0])  + '\n'
        string += f'{" "*4}{matrix_name}_nja = ' + str(spmatrix_list[3][1])  + '\n'
        string += f'{" "*4}return {matrix_name}_ia, {matrix_name}_ja, {matrix_name}_nia, {matrix_name}_nja \n'
        
        return string


    def sys2num(self):
        sys = self.sys

        t_0 = time.time()
        
        name = sys['name']
        N_x = sys['N_x']
        N_y = sys['N_y']
        N_z = sys['N_z']
        N_u = sys['N_u']



        class_template = pkgutil.get_data(__name__, "templates/class_dae_cffi_template.py").decode().replace('\r\n','\n') 
        functions_template = pkgutil.get_data(__name__, "templates/functions_template.py").decode().replace('\r\n','\n') 
    #     solver_template = pkgutil.get_data(__name__, "templates/solver_template_v2.py").decode().replace('\r\n','\n') 

        class_template = class_template.replace('{name}',str(name))
        class_template = class_template.replace('{u2z_jacobians}',self.string_u2z)
        class_template = class_template.replace('{u2z_comment}',self.u2z_comment)

        class_template = class_template.replace('{N_x}',str(N_x)).replace('{N_y}',str(N_y)).replace('{N_z}',str(N_z))
        class_template = class_template.replace('{x_list}', str([str(item) for item in sys['x']]))
        class_template = class_template.replace('{y_run_list}', str([str(item) for item in sys['y_run']]))
        class_template = class_template.replace('{y_ini_list}', str([str(item) for item in sys['y_ini']]))
        class_template = class_template.replace('{params_list}', str([str(item) for item in sys['params_dict']]))
        class_template = class_template.replace('{params_values_list}', str([(sys['params_dict'][item]) for item in sys['params_dict']]))
        class_template = class_template.replace('{inputs_ini_list}', str([str(item) for item in sys['u_ini']]))
        class_template = class_template.replace('{inputs_run_list}', str([str(item) for item in sys['u_run']]))
        class_template = class_template.replace('{inputs_ini_values_list}', str([(sys['u_ini_dict'][item]) for item in sys['u_ini_dict']]))
        class_template = class_template.replace('{inputs_run_values_list}', str([(sys['u_run_dict'][item]) for item in sys['u_run_dict']]))
        class_template = class_template.replace('{outputs_list}', str([str(item) for item in sys['h_dict']]))


        if 'enviroment' in sys:
            class_template = class_template.replace('{enviroment_name}',str(sys['enviroment']))
            class_template = class_template.replace('{dae_file_mode}',"'enviroment'")
        elif 'colab' in sys:
            class_template = class_template.replace('{enviroment_name}',str(sys['colab']))
            class_template = class_template.replace('{dae_file_mode}',"'colab'")      
        elif 'testing' in sys:
            class_template = class_template.replace('{enviroment_name}','no_enviroment')
            class_template = class_template.replace('{dae_file_mode}',"'testing'")        
        else:
            class_template = class_template.replace('{enviroment_name}','no_enviroment')
            class_template = class_template.replace('{dae_file_mode}',"'local'")
            
            
        module = class_template
        
        module += '\n'*2
        module += self.spmatrix2def(sys['sp_jac_ini_list'],'sp_jac_ini')
        module += '\n'
        module += self.spmatrix2def(sys['sp_jac_run_list'],'sp_jac_run')
        module += '\n'
        module += self.spmatrix2def(sys['sp_jac_trap_list'],'sp_jac_trap')
        
        

           
        with open(f'{name}.py','w') as fobj:
            fobj.write(module)
    
        if self.verbose: print(f'sys2num (time: {time.time()-t_0:0.3f})')                 
            




    def sym2xyup(self,string,sys,inirun):
        i = 0
        for item in sys['x']:
            string = re.sub(f"\\b{item}\\b"  ,f'x[{i}]',string)
            i+=1

        i = 0
        for item in sys[f'y_{inirun}']:
            string = re.sub(f"\\b{item}\\b"  ,f'y[{i}]',string)
            i+=1

        i = 0
        for item in sys[f'u_{inirun}']:
            string = re.sub(f"\\b{item}\\b"  ,f'u[{i}]',string)
            i+=1

        i = 0
        for item in sys['params_dict']:
            string = re.sub(f"\\b{item}\\b"  ,f'p[{i}]',string)
            i+=1
            
        return string

    def sym2str(self,fun_name,vector,sys,inirun):
        LHS = sym.Symbol('LHS')

        string_sym = ''
        for irow in range(len(vector)):
            string_sym += sym.ccode(Assignment(LHS,vector[irow])).replace('LHS',f'out[{irow}]') +'\n#'
            
        string_src = self.sym2xyup(string_sym,sys,inirun=inirun)
        string_split = string_src.split('#')
        string = ''
        for irow in range(len(vector)):
            string_sym = sym.ccode(Assignment(LHS,vector[irow])).replace('LHS',f'out[{irow}]') +'\n'
            string += string_split[irow]

        defs   = f'void {fun_name}_eval(double *out,double *x,double *y,double *u,double *p,double Dt);' + '\n' 
        source = f'void {fun_name}_eval(double *out,double *x,double *y,double *u,double *p,double Dt)'  + '{'+ '\n'*2
        source += string + '\n}\n'

        self.defs = defs
        self.source = source
        return defs,source
            
            
                    
    def sym2rhs(self,data,indices,indptr,shape,sys,inirun):
        '''
        Takes a sparse symbolic CRS matrix (data,indices,indptr) and convert it as a
        string replacing symbolic names by x,y,u,p names.
        
        
        '''

        
        rhs_list = []
        
        for irow in range(len(indptr)-1):
            for k in range(indptr[irow],indptr[irow+1]):
                icol = indices[k]
                data_k = data[k]
                if not data[irow] == 0:
                    string_sym = sym.ccode(data_k) + ';\n'
                    string_i = self.sym2xyup(string_sym,sys,inirun)

                    tipo = 'num'
                    if 'x[' in string_i or 'y[' in string_i:
                        tipo = 'xy' 
                    elif 'p[' in string_i or 'u[' in string_i or 'Dt' in string_i:
                        tipo = 'up' 
                
                    rhs_list += [(string_i,tipo,irow,icol)]
                    
        return rhs_list

    def sym2rhs2(self,data,indices,indptr,shape,sys,inirun):


        rhs_list = []
        
        string_full_sym = ''
        
        for irow in range(len(indptr)-1):
            for k in range(indptr[irow],indptr[irow+1]):
                icol = indices[k]
                data_k = data[k]
                if not data_k == 0:
                    string_sym = sym.ccode(sym.N(data_k)) + ';\n#'
                    string_full_sym += string_sym
        

        string_full = self.sym2xyup(string_full_sym,sys,inirun)
        
        string_split = string_full.split('#')  
        for irow in range(len(indptr)-1):
            for k in range(indptr[irow],indptr[irow+1]):
                icol = indices[k]
                data_k = string_split[k]
                if not data_k == 0:
                    string_i = data_k
                    
                    tipo = 'num'
                    if 'x[' in string_i or 'y[' in string_i:
                        tipo = 'xy' 
                    elif 'p[' in string_i or 'u[' in string_i or 'Dt' in string_i:
                        tipo = 'up' 
                
                    rhs_list += [(string_i,tipo,irow,icol)]

                    
        return rhs_list


    def rhs2str(self,rhs_list,lhs_name,num_matrix,shape,mode='crs'):
        string_xy = ''
        string_up = ''
        string_num = ''
        N_col = shape[1]

        k = 0    
        for data_i,tipo,irow,icol in rhs_list:
            if mode == 'crs':
                idx = k
            if mode == 'dense':
                idx = irow*N_col+icol
            if mode == '2d':
                idx = f'{irow},{icol}'
                
            if tipo == 'xy':
                string_xy += f'{lhs_name}[{idx}] = {data_i}'
            if tipo == 'up':
                string_up += f'{lhs_name}[{idx}] = {data_i}'
            if tipo == 'num':
                string_num += f'{lhs_name}[{idx}] = {data_i}'  
                num_matrix[k] = float(data_i[:-2])
            k+=1
                    
        return string_xy,string_up,string_num
            

    def str2src(self,fun_name,string_xy,string_up,string_num,matrix_name='out'):
    

        defs = ''
        defs += f'void {fun_name}_xy_eval(double *{matrix_name},double *x,double *y,double *u,double *p,double Dt);' + '\n' 
        defs += f'void {fun_name}_up_eval(double *{matrix_name},double *x,double *y,double *u,double *p,double Dt);' + '\n' 
        defs += f'void {fun_name}_num_eval(double *{matrix_name},double *x,double *y,double *u,double *p,double Dt);' + '\n' 
            
        source = '' 
        source += f'void {fun_name}_xy_eval(double *{matrix_name},double *x,double *y,double *u,double *p,double Dt)' + '{' +'\n'*2
        source += string_xy
        source += '\n}\n\n'
        source += f'void {fun_name}_up_eval(double *{matrix_name},double *x,double *y,double *u,double *p,double Dt)'  + '{'+ '\n'*2
        source += string_up
        source += '\n}\n\n'
        source += f'void {fun_name}_num_eval(double *{matrix_name},double *x,double *y,double *u,double *p,double Dt)'  + '{'+ '\n'*2
        source += string_num
        source += '\n}\n\n'
        
        return defs,source
        

    def sym2src(self):
        sys = self.sys
        
        t_0 = time.time()
        
        ## jac_run ################################################################
        if self.verbose: print(f'writting f_ini and g_ini code (time: {time.time()-t_0:0.3f} s)')

        defs_f_ini,source_f_ini = self.sym2str('f_ini',sys['f'],sys,'ini')
        defs_g_ini,source_g_ini = self.sym2str('g_ini',sys['g'],sys,'ini')

        if self.verbose: print(f'writting f_run and g_run code (time: {time.time()-t_0:0.3f} s)')
        
        defs_f_run,source_f_run = self.sym2str('f_run',sys['f'],sys,'run')
        defs_g_run,source_g_run = self.sym2str('g_run',sys['g'],sys,'run')

        if self.verbose: print(f'writting h_run code (time: {time.time()-t_0:0.3f} s)')   
        defs_h,source_h = self.sym2str('h',sys['h'],sys,'run')
        
        ## jac_ini ################################################################
        if self.verbose: print(f'converting jac_ini to sp_jac_ini  (time: {time.time()-t_0:0.3f} s)')
        
        jac_ini=sys['jac_ini']

        sp_jac_ini_list = _doktocsr(SparseMatrix(jac_ini))
        sys['sp_jac_ini_list'] = sp_jac_ini_list

        
        data = sp_jac_ini_list[0]
        indices = sp_jac_ini_list[1]
        indptr = sp_jac_ini_list[2]
        shape = sp_jac_ini_list[3]
        
        sp_jac_ini_num_matrix = csr_matrix((np.zeros(len(data)),indices,indptr),shape=shape)
        
        if self.verbose: print(f'running sym2rhs for sp_jac_ini (time: {time.time()-t_0:0.3f} s)')

        rhs_list = self.sym2rhs2(data,indices,indptr,shape,sys,'ini')   
        
        string_xy,string_up,string_num = self.rhs2str(rhs_list,'out',sp_jac_ini_num_matrix.data,shape,mode='dense')  
        
        if self.jac_num:
            defs_de_ini,source_de_ini = self.str2src('de_jac_ini',string_xy,string_up,string_num,matrix_name='out')
        else:
            defs_de_ini,source_de_ini = self.str2src('de_jac_ini',string_xy,string_up,'\n',matrix_name='out')
            

        string_xy,string_up,string_num = self.rhs2str(rhs_list,'out',sp_jac_ini_num_matrix.data,shape,mode='crs')
        
        if self.jac_num:
            defs_sp_ini,source_sp_ini = self.str2src('sp_jac_ini',string_xy,string_up,string_num,matrix_name='out')
        else:
            defs_sp_ini,source_sp_ini = self.str2src('sp_jac_ini',string_xy,string_up,'\n',matrix_name='out')

               
        ## jac_run ################################################################
        if self.verbose: print(f'converting jac_run to sp_jac_run  (time: {time.time()-t_0:0.3f} s)')
        
        jac_run=sys['jac_run']

        sp_jac_run_list = _doktocsr(SparseMatrix(jac_run))
        
        data = sp_jac_run_list[0]
        indices = sp_jac_run_list[1]
        indptr = sp_jac_run_list[2]
        shape = sp_jac_run_list[3]
        
        sp_jac_run_num_matrix = csr_matrix((np.zeros(len(data)),indices,indptr),shape=shape)

        
        if self.verbose: print(f'running sym2rhs for sp_jac_run (time: {time.time()-t_0:0.3f} s)')
        #sym2rhs(data,indices,indptr,shape,sys,inirun)
        #rhs_list = sym2rhs2(data,indices,indptr,shape,sys,'run')       


        sys['sp_jac_run_list'] = sp_jac_run_list
        
        data = sp_jac_run_list[0]
        indices = sp_jac_run_list[1]
        indptr = sp_jac_run_list[2]
        shape = sp_jac_run_list[3]
        
        rhs_list = self.sym2rhs2(data,indices,indptr,shape,sys,'run')   
        
        
        string_xy,string_up,string_num = self.rhs2str(rhs_list,'out',sp_jac_run_num_matrix.data,shape,mode='dense')
        
        if self.jac_num:  
            defs_de_run,source_de_run = self.str2src('de_jac_run',string_xy,string_up,string_num,matrix_name='out')
        else:
            defs_de_run,source_de_run = self.str2src('de_jac_run',string_xy,string_up,'\n',matrix_name='out')

        
        string_xy,string_up,string_num = self.rhs2str(rhs_list,'out',sp_jac_run_num_matrix.data,shape,mode='crs')
        if self.jac_num:
            defs_sp_run,source_sp_run = self.str2src('sp_jac_run',string_xy,string_up,string_num,matrix_name='out')
        else:    
            defs_sp_run,source_sp_run = self.str2src('sp_jac_run',string_xy,string_up,'\n',matrix_name='out')
        
        
        ## jac_trap ###############################################################
        if self.verbose: print(f'converting jac_trap to sp_jac_trap  (time: {time.time()-t_0:0.3f} s)')  
        
        jac_trap=sys['jac_trap']

        sp_jac_trap_list = _doktocsr(SparseMatrix(jac_trap))

        sys['sp_jac_trap_list'] = sp_jac_trap_list
        data = sp_jac_trap_list[0]
        indices = sp_jac_trap_list[1]
        indptr = sp_jac_trap_list[2]
        
        sp_jac_trap_num_matrix = csr_matrix((np.zeros(len(data)),indices,indptr),shape=shape)

        
        if self.verbose: print(f'running sym2rhs for sp_jac_trap (time: {time.time()-t_0:0.3f} s)')
        rhs_list = self.sym2rhs2(data,indices,indptr,shape,sys,'run')       

        if self.verbose: print(f'wrtting  de_jac_trap code (time: {time.time()-t_0:0.3f} s)')   
        string_xy,string_up,string_num = self.rhs2str(rhs_list,'out',sp_jac_trap_num_matrix.data,shape,mode='dense')
        
        if self.jac_num:
            defs_de_trap,source_de_trap = self.str2src('de_jac_trap',string_xy,string_up,string_num,matrix_name='out')
        else:
            defs_de_trap,source_de_trap = self.str2src('de_jac_trap',string_xy,string_up,'\n',matrix_name='out')
            

        if self.verbose: print(f'writting sp_jac_trap code (time: {time.time()-t_0:0.3f} s)')      
        string_xy,string_up,string_num = self.rhs2str(rhs_list,'out',sp_jac_trap_num_matrix.data,shape,mode='crs')
        if self.jac_num:
            defs_sp_trap,source_sp_trap = self.str2src('sp_jac_trap',string_xy,string_up,string_num,matrix_name='out')
        else:
            defs_sp_trap,source_sp_trap = self.str2src('sp_jac_trap',string_xy,string_up,'\n',matrix_name='out')
            

        ## Fu_run and Gu_run ###############################################################
        defs_de_uz   = ''
        source_de_uz = ''
        defs_sp_uz   = ''
        source_sp_uz = ''

        if not 'uz_jacs' in sys:
            self.uz_jacs = False
            self.string_u2z = '\n'
            self.u2z_comment = '#'
        else:
            self.uz_jacs = sys['uz_jacs'] 
            self.u2z_comment = ''


        if self.uz_jacs:
            self.string_u2z = 'sp_Fu_run_up_eval = jacs.lib.sp_Fu_run_up_eval\n'
            self.string_u2z+= 'sp_Gu_run_up_eval = jacs.lib.sp_Gu_run_up_eval\n'
            self.string_u2z+= 'sp_Hx_run_up_eval = jacs.lib.sp_Hx_run_up_eval\n'
            self.string_u2z+= 'sp_Hy_run_up_eval = jacs.lib.sp_Hy_run_up_eval\n'
            self.string_u2z+= 'sp_Hu_run_up_eval = jacs.lib.sp_Hu_run_up_eval\n'
            self.string_u2z+= 'sp_Fu_run_xy_eval = jacs.lib.sp_Fu_run_xy_eval\n'
            self.string_u2z+= 'sp_Gu_run_xy_eval = jacs.lib.sp_Gu_run_xy_eval\n'
            self.string_u2z+= 'sp_Hx_run_xy_eval = jacs.lib.sp_Hx_run_xy_eval\n'
            self.string_u2z+= 'sp_Hy_run_xy_eval = jacs.lib.sp_Hy_run_xy_eval\n'
            self.string_u2z+= 'sp_Hu_run_xy_eval = jacs.lib.sp_Hu_run_xy_eval\n'

            for item in ['Fu_run','Gu_run','Hx_run','Hy_run','Hu_run']:
                if self.verbose: print(f'converting {item} to sparse (time: {time.time()-t_0:0.3f} s)')  
                sp_jac_list = _doktocsr(SparseMatrix(sys[item]))  
                data = sp_jac_list[0]
                indices = sp_jac_list[1]
                indptr = sp_jac_list[2]
                shape = sp_jac_list[3]
            
                sp_jac_num_matrix = csr_matrix((np.zeros(len(data)),indices,indptr),shape=shape)
            
            
                if self.verbose: print(f'running sym2rhs for {item} (time: {time.time()-t_0:0.3f} s)')
                rhs_list = self.sym2rhs2(data,indices,indptr,shape,sys,'run')       

                if self.verbose: print(f'writting  {item} code (time: {time.time()-t_0:0.3f} s)')   
                string_xy,string_up,string_num = self.rhs2str(rhs_list,'out',sp_jac_num_matrix.data,shape,mode='dense')
            
                if self.jac_num:
                    defs_de_jac,source_de_jac = self.str2src(f'de_{item}',string_xy,string_up,string_num,matrix_name='out')
                else:
                    defs_de_jac,source_de_jac = self.str2src(f'de_{item}',string_xy,string_up,'\n',matrix_name='out')
                    

                if self.verbose: print(f'writting {item} code (time: {time.time()-t_0:0.3f} s)')      
                string_xy,string_up,string_num = self.rhs2str(rhs_list,'out',sp_jac_num_matrix.data,shape,mode='crs')
                if self.jac_num:
                    defs_sp_jac,source_sp_jac = self.str2src(f'sp_{item}',string_xy,string_up,string_num,matrix_name='out')
                else:
                    defs_sp_jac,source_sp_jac = self.str2src(f'sp_{item}',string_xy,string_up,'\n',matrix_name='out')

                defs_de_uz   += defs_de_jac
                source_de_uz += source_de_jac
                defs_sp_uz   += defs_sp_jac
                source_sp_uz += source_sp_jac
                
                if self.matrices_folder == '':
                    save_npz(f"./{sys['name']}_{item}_num.npz",sp_jac_num_matrix, compressed=True)
                else:
                    save_npz(f"./{self.matrices_folder}/{sys['name']}_{item}_num.npz",sp_jac_num_matrix, compressed=True)


        ## C sources ##############################################################
        if self.verbose: print(f'writting full source (time: {time.time()-t_0:0.3f} s)')   
        
        defs = defs_f_ini + defs_g_ini 
        defs += defs_f_run + defs_g_run  
        defs += defs_h  
        defs += defs_de_ini + defs_sp_ini 
        defs += defs_de_run + defs_sp_run 
        defs += defs_de_trap + defs_sp_trap 
        defs += defs_de_uz + defs_sp_uz
            
        source = source_f_ini + source_g_ini 
        source += source_f_run + source_g_run
        source += source_h
        source += source_de_ini + source_sp_ini 
        source += source_de_run + source_sp_run
        source += source_de_trap + source_sp_trap
        source += source_de_uz + source_sp_uz

        save_npz( f"./{self.matrices_folder}/{sys['name']}_sp_jac_ini_num.npz", sp_jac_ini_num_matrix, compressed=True)
        save_npz( f"./{self.matrices_folder}/{sys['name']}_sp_jac_run_num.npz", sp_jac_run_num_matrix, compressed=True)
        save_npz( f"./{self.matrices_folder}/{sys['name']}_sp_jac_trap_num.npz",sp_jac_trap_num_matrix, compressed=True)

        if self.verbose: print(f'Code wrote in {time.time()-t_0:0.3f} s')

        self.defs = defs
        self.source = source

        return defs,source

    def compile_module(self):
        name = self.name
        defs = self.defs
        source = self.source
        module_name = self.name + '_cffi'
        
        ffi = cffi.FFI()
        ffi.cdef(defs, override=True)
        ffi.set_source(module_name=module_name,source=source)
        t_0 = time.time()
        ffi.compile()
        if self.verbose: print(f'Compilation time: {time.time()-t_0:0.2f} s')

    def compile_module_files(self):
        
        ffi = cffi.FFI()
        name = self.name

        with open(f'./build/defs_{name}_cffi.h', 'r') as f:
            defs = f.read()
        with open(f'./build/source_{name}_cffi.c', 'r') as f:
            source = f.read()
            
        ffi.cdef(defs, override=True)
        ffi.set_source(module_name=f"{name}",source=source)
        t_0 = time.time()
        ffi.compile()
        print(f'Compilation time: {time.time()-t_0:0.2f} s')
        
    def build(self):

        sys = self.sys
        name = self.name
        sys = self.system()
        self.jacobians()
        defs,source = self.sym2src()
        if self.save_sources:
            with open(f'./build/defs_{name}_cffi.h', 'w') as f:
                f.write(defs)
            with open(f'./build/source_{name}_cffi.c', 'w') as f:
                f.write(source)
        self.compile_module()
        self.sys2num()
            
    
if __name__ == "__main__":

    
    params_dict = {'X_d':1.81,'X1d':0.3, 'T1d0':8.0,  # synnchronous machine d-axis parameters
                   'X_q':1.76,'X1q':0.65,'T1q0':1.0,  # synnchronous machine q-axis parameters
                   'R_a':0.003,
                   'X_l': 0.02, 
                   'H':3.5,'D':0.0,
                   'Omega_b':2*np.pi*50,'omega_s':1.0,
                   'V_0':1.0,'Theta_0':0.0}
    
    
    u_ini_dict = {'p_m':0.8,'v_f':1.0}  # for the initialization problem
    u_run_dict = {'p_m':0.8,'v_f':1.0}  # for the running problem (here initialization and running problem are the same)
    
    
    x_list = ['delta','omega','e1q','e1d']    # dynamic states
    y_ini_list = ['i_d','i_q','p_t','q_t','v_t','theta_t']   
    y_run_list = ['i_d','i_q','p_t','q_t','v_t','theta_t']   
    
    sys_vars = {'params':params_dict,
                'u_list':u_run_dict,
                'x_list':x_list,
                'y_list':y_run_list}
    
    exec(sym_gen_str())  # exec to generate the required symbolic varables and constants

    # auxiliar equations
    v_d = v_t*sin(delta - theta_t)  # park
    v_q = v_t*cos(delta - theta_t)  # park
    
    p_e = i_d*(v_d + R_a*i_d) + i_q*(v_q + R_a*i_q) # electromagnetic power
    
    # dynamic equations
    ddelta = Omega_b*(omega - omega_s) - 1e-4*delta  # load angle
    domega = 1/(2*H)*(p_m - p_e - D*(omega - omega_s)) # speed
    de1q = 1/T1d0*(-e1q - (X_d - X1d)*i_d + v_f)
    de1d = 1/T1q0*(-e1d + (X_q - X1q)*i_q)
    
    # algebraic equations
    g_1 = v_q + R_a*i_q + X1d*i_d - e1q # stator
    g_2 = v_d + R_a*i_d - X1q*i_q - e1d # stator
    g_3 = i_d*v_d + i_q*v_q - p_t # active power 
    g_4 = i_d*v_q - i_q*v_d - q_t # reactive power
    g_5 = p_t - (v_t*V_0*sin(theta_t - Theta_0))/X_l  # network equation (p)
    g_6 = q_t + (v_t*V_0*cos(theta_t - Theta_0))/X_l - v_t**2/X_l  # network equation (q) 

    sys_dict = {'name':'build_test',
           'params_dict':params_dict,
           'f_list':[ddelta,domega,de1q,de1d],
           'g_list':[g_1,g_2,g_3,g_4,g_5,g_6],
           'x_list':x_list,
           'y_ini_list':y_ini_list,
           'y_run_list':y_run_list,
           'u_run_dict':u_run_dict,
           'u_ini_dict':u_ini_dict,
           'h_dict':{'p_m':p_m,'p_e':p_e, 'v_f':v_f}}
    
    bd = builder(sys_dict)
    bd.build()
    #sys2num('build_test',sys, verbose=True)
   # sys2num(sys_dict)                                                             


    
